from pydantic import BaseModel, StrictStr, StrictInt, Field
from typing import Literal

class RDSDBProperties(BaseModel):
    AvailabilityZone: StrictStr
    BackupRetentionPeriod: StrictInt
    DBInstanceClass: StrictStr = Field(pattern=r'^((db)\.(c6g|c6gd|c6gn|c6i|c6id|c7g|g5g|im4gn|is4gen|m6g|m6gd|r6g|r6gd|t4g|x2gd)\.(10xlarge|112xlarge|12xlarge|16xlarge|18xlarge|24xlarge|2xlarge|32xlarge|3xlarge|48xlarge|4xlarge|56xlarge|6xlarge|8xlarge|9xlarge|large|medium|metal|micro|nano|small|xlarge))')
    StorageType: Literal['sc1', 'st1', 'gp3']

class RDSDBInstance(BaseModel):
    Type: Literal["AWS::RDS::DBInstance"]

class EC2EbsDevice(BaseModel):
    VolumeType: StrictStr
    DeleteOnTermination: Literal[True]

class EC2BlockDeviceMappings(BaseModel):
    Ebs: EC2EbsDevice

class EC2VolumeProperties(BaseModel):
    VolumeType: Literal["sc1", "st1", "gp3"]
    BlockDeviceMappings: EC2BlockDeviceMappings

class EC2Volume(BaseModel):
    Type: Literal["AWS::EC2::Volume"]
    Properties: EC2VolumeProperties

class EC2InstanceProperties(BaseModel):
    InstanceType: StrictStr = Field(pattern=r'^((c6g|c6gd|c6gn|c6i|c6id|c7g|g5g|im4gn|is4gen|m6g|m6gd|r6g|r6gd|t4g|x2gd)\.(10xlarge|112xlarge|12xlarge|16xlarge|18xlarge|24xlarge|2xlarge|32xlarge|3xlarge|48xlarge|4xlarge|56xlarge|6xlarge|8xlarge|9xlarge|large|medium|metal|micro|nano|small|xlarge))')

class EC2Instance(BaseModel):
    Type: Literal["AWS::EC2::Instance"]
    Properties: EC2InstanceProperties

class LoggingGroupProperties(BaseModel):
    LogGroupClass: Literal["Infrequent Access"]
    RetentionInDays: StrictInt

class LoggingGroup(BaseModel):
    Type: Literal["AWS::Logs::LogGroup"]
    Properties: LoggingGroupProperties

class LambdaLoggingConfig(BaseModel):
    LogGroup: StrictStr

class LambdaProperties(BaseModel):
    LoggingConfig: LambdaLoggingConfig

class Lambda(BaseModel):
    Type: Literal["AWS::Serverless::Function", "AWS::Lambda::Function"]
    Properties: LambdaProperties

class Resource(BaseModel):
    Type: StrictStr



from __future__ import annotations

# This module contains abstractions for the input stream. You don't have to
# looks further, there are no pretty code.
#
# We define two classes here.
#
#   Mark(source, line, column)
# It's just a record and its only use is producing nice error messages.
# Parser does not use it for any other purposes.
#
#   Reader(source, data)
# Reader determines the encoding of `data` and converts it to unicode.
# Reader provides the following methods and attributes:
#   reader.peek(length=1) - return the next `length` characters
#   reader.forward(length=1) - move the current position to `length`
#      characters.
#   reader.index - the number of the current character.
#   reader.line, stream.column - the line and the column of the current
#      character.

import codecs

from cfn_check.yaml.error import YAMLError, FileMark, StringMark, YAMLStreamError
from cfn_check.yaml.util import RegExp
from typing import Any, Dict, Optional, List, Union, Text, Tuple, Optional  # NOQA

__all__ = ['Reader', 'ReaderError']


class ReaderError(YAMLError):
    def __init__(
        self, name: Any, position: Any, character: Any, encoding: Any, reason: Any,
    ) -> None:
        self.name = name
        self.character = character
        self.position = position
        self.encoding = encoding
        self.reason = reason

    def __str__(self) -> Any:
        if isinstance(self.character, bytes):
            return (
                f"'{self.encoding!s}' codec can't decode byte #x{ord(self.character):02x}: "
                f'{self.reason!s}\n'
                f'  in "{self.name!s}", position {self.position:d}'
            )
        else:
            return (
                f'unacceptable character #x{self.character:04x}: {self.reason!s}\n'
                f'  in "{self.name!s}", position {self.position:d}'
            )


class Reader:
    # Reader:
    # - determines the data encoding and converts it to a unicode string,
    # - checks if characters are in allowed range,
    # - adds '\0' to the end.

    # Reader accepts
    #  - a `bytes` object,
    #  - a `str` object,
    #  - a file-like object with its `read` method returning `str`,
    #  - a file-like object with its `read` method returning `unicode`.

    # Yeah, it's ugly and slow.

    def __init__(self, stream: Any, loader: Any = None) -> None:
        self.loader = loader
        if self.loader is not None and getattr(self.loader, '_reader', None) is None:
            self.loader._reader = self
        self.reset_reader()
        self.stream: Any = stream  # as .read is called

    def reset_reader(self) -> None:
        self.name: Any = None
        self.stream_pointer = 0
        self.eof = True
        self.buffer = ""
        self.pointer = 0
        self.raw_buffer: Any = None
        self.raw_decode = None
        self.encoding: Optional[Text] = None
        self.index = 0
        self.line = 0
        self.column = 0

    @property
    def stream(self) -> Any:
        try:
            return self._stream
        except AttributeError:
            raise YAMLStreamError('input stream needs to be specified')

    @stream.setter
    def stream(self, val: Any) -> None:
        if val is None:
            return
        self._stream = None
        if isinstance(val, str):
            self.name = '<unicode string>'
            self.check_printable(val)
            self.buffer = val + '\0'
        elif isinstance(val, bytes):
            self.name = '<byte string>'
            self.raw_buffer = val
            self.determine_encoding()
        else:
            if not hasattr(val, 'read'):
                raise YAMLStreamError('stream argument needs to have a read() method')
            self._stream = val
            self.name = getattr(self.stream, 'name', '<file>')
            self.eof = False
            self.raw_buffer = None
            self.determine_encoding()

    def peek(self, index: int = 0) -> Text:
        try:
            return self.buffer[self.pointer + index]
        except IndexError:
            self.update(index + 1)
            return self.buffer[self.pointer + index]

    def prefix(self, length: int = 1) -> Any:
        if self.pointer + length >= len(self.buffer):
            self.update(length)
        return self.buffer[self.pointer : self.pointer + length]

    def forward_1_1(self, length: int = 1) -> None:
        if self.pointer + length + 1 >= len(self.buffer):
            self.update(length + 1)
        while length != 0:
            ch = self.buffer[self.pointer]
            self.pointer += 1
            self.index += 1
            if ch in '\n\x85\u2028\u2029' or (
                ch == '\r' and self.buffer[self.pointer] != '\n'
            ):
                self.line += 1
                self.column = 0
            elif ch != '\uFEFF':
                self.column += 1
            length -= 1

    def forward(self, length: int = 1) -> None:
        if self.pointer + length + 1 >= len(self.buffer):
            self.update(length + 1)
        while length != 0:
            ch = self.buffer[self.pointer]
            self.pointer += 1
            self.index += 1
            if ch == '\n' or (ch == '\r' and self.buffer[self.pointer] != '\n'):
                self.line += 1
                self.column = 0
            elif ch != '\uFEFF':
                self.column += 1
            length -= 1

    def get_mark(self) -> Any:
        if self.stream is None:
            return StringMark(
                self.name, self.index, self.line, self.column, self.buffer, self.pointer,
            )
        else:
            return FileMark(self.name, self.index, self.line, self.column)

    def determine_encoding(self) -> None:
        while not self.eof and (self.raw_buffer is None or len(self.raw_buffer) < 2):
            self.update_raw()
        if isinstance(self.raw_buffer, bytes):
            if self.raw_buffer.startswith(codecs.BOM_UTF16_LE):
                self.raw_decode = codecs.utf_16_le_decode  # type: ignore
                self.encoding = 'utf-16-le'
            elif self.raw_buffer.startswith(codecs.BOM_UTF16_BE):
                self.raw_decode = codecs.utf_16_be_decode  # type: ignore
                self.encoding = 'utf-16-be'
            else:
                self.raw_decode = codecs.utf_8_decode  # type: ignore
                self.encoding = 'utf-8'
        self.update(1)

    NON_PRINTABLE = RegExp(
        '[^\x09\x0A\x0D\x20-\x7E\x85' '\xA0-\uD7FF' '\uE000-\uFFFD' '\U00010000-\U0010FFFF' ']'  # NOQA
    )

    _printable_ascii = ('\x09\x0A\x0D' + "".join(map(chr, range(0x20, 0x7F)))).encode('ascii')

    @classmethod
    def _get_non_printable_ascii(cls: Text, data: bytes) -> Optional[Tuple[int, Text]]:  # type: ignore # NOQA
        ascii_bytes = data.encode('ascii')  # type: ignore
        non_printables = ascii_bytes.translate(None, cls._printable_ascii)  # type: ignore
        if not non_printables:
            return None
        non_printable = non_printables[:1]
        return ascii_bytes.index(non_printable), non_printable.decode('ascii')

    @classmethod
    def _get_non_printable_regex(cls, data: Text) -> Optional[Tuple[int, Text]]:
        match = cls.NON_PRINTABLE.search(data)
        if not bool(match):
            return None
        return match.start(), match.group()

    @classmethod
    def _get_non_printable(cls, data: Text) -> Optional[Tuple[int, Text]]:
        try:
            return cls._get_non_printable_ascii(data)  # type: ignore
        except UnicodeEncodeError:
            return cls._get_non_printable_regex(data)

    def check_printable(self, data: Any) -> None:
        non_printable_match = self._get_non_printable(data)
        if non_printable_match is not None:
            start, character = non_printable_match
            position = self.index + (len(self.buffer) - self.pointer) + start
            raise ReaderError(
                self.name,
                position,
                ord(character),
                'unicode',
                'special characters are not allowed',
            )

    def update(self, length: int) -> None:
        if self.raw_buffer is None:
            return
        self.buffer = self.buffer[self.pointer :]
        self.pointer = 0
        while len(self.buffer) < length:
            if not self.eof:
                self.update_raw()
            if self.raw_decode is not None:
                try:
                    data, converted = self.raw_decode(self.raw_buffer, 'strict', self.eof)
                except UnicodeDecodeError as exc:
                    character = self.raw_buffer[exc.start]
                    if self.stream is not None:
                        position = self.stream_pointer - len(self.raw_buffer) + exc.start
                    elif self.stream is not None:
                        position = self.stream_pointer - len(self.raw_buffer) + exc.start
                    else:
                        position = exc.start
                    raise ReaderError(self.name, position, character, exc.encoding, exc.reason)
            else:
                data = self.raw_buffer
                converted = len(data)
            self.check_printable(data)
            self.buffer += data
            self.raw_buffer = self.raw_buffer[converted:]
            if self.eof:
                self.buffer += '\0'
                self.raw_buffer = None
                break

    def update_raw(self, size: Optional[int] = None) -> None:
        if size is None:
            size = 4096
        data = self.stream.read(size)
        if self.raw_buffer is None:
            self.raw_buffer = data
        else:
            self.raw_buffer += data
        self.stream_pointer += len(data)
        if not data:
            self.eof = True


# try:
#     import psyco
#     psyco.bind(Reader)
# except ImportError:
#     pass




from __future__ import annotations

from cfn_check.yaml.compat import nprintf  # NOQA
from typing import Text, Any, Dict, Optional, List  # NOQA
from .error import StreamMark  # NOQA

SHOW_LINES = True


class Token:
    __slots__ = 'start_mark', 'end_mark', '_comment'

    def __init__(self, start_mark: StreamMark, end_mark: StreamMark) -> None:
        self.start_mark = start_mark
        self.end_mark = end_mark

    def __repr__(self) -> Any:
        # attributes = [key for key in self.__slots__ if not key.endswith('_mark') and
        #               hasattr('self', key)]
        attributes = [key for key in self.__slots__ if not key.endswith('_mark')]
        attributes.sort()
        # arguments = ', '.join(
        #  [f'{key!s}={getattr(self, key)!r})' for key in attributes]
        # )
        arguments = [f'{key!s}={getattr(self, key)!r}' for key in attributes]
        if SHOW_LINES:
            try:
                arguments.append('line: ' + str(self.start_mark.line))
            except:  # NOQA
                pass
        try:
            arguments.append('comment: ' + str(self._comment))
        except:  # NOQA
            pass
        return f'{self.__class__.__name__}({", ".join(arguments)})'

    @property
    def column(self) -> int:
        return self.start_mark.column

    @column.setter
    def column(self, pos: Any) -> None:
        self.start_mark.column = pos

    # old style ( <= 0.17) is a TWO element list with first being the EOL
    # comment concatenated with following FLC/BLNK; and second being a list of FLC/BLNK
    # preceding the token
    # new style ( >= 0.17 ) is a THREE element list with the first being a list of
    # preceding FLC/BLNK, the second EOL and the third following FLC/BLNK
    # note that new style has differing order, and does not consist of CommentToken(s)
    # but of CommentInfo instances
    # any non-assigned values in new style are None, but first and last can be empty list
    # new style routines add one comment at a time

    # going to be deprecated in favour of add_comment_eol/post
    def add_post_comment(self, comment: Any) -> None:
        if not hasattr(self, '_comment'):
            self._comment = [None, None]
        else:
            assert len(self._comment) in [2, 5]  # make sure it is version 0
        # if isinstance(comment, CommentToken):
        #    if comment.value.startswith('# C09'):
        #        raise
        self._comment[0] = comment

    # going to be deprecated in favour of add_comment_pre
    def add_pre_comments(self, comments: Any) -> None:
        if not hasattr(self, '_comment'):
            self._comment = [None, None]
        else:
            assert len(self._comment) == 2  # make sure it is version 0
        assert self._comment[1] is None
        self._comment[1] = comments
        return

    # new style
    def add_comment_pre(self, comment: Any) -> None:
        if not hasattr(self, '_comment'):
            self._comment = [[], None, None]  # type: ignore
        else:
            assert len(self._comment) == 3
            if self._comment[0] is None:
                self._comment[0] = []  # type: ignore
        self._comment[0].append(comment)  # type: ignore

    def add_comment_eol(self, comment: Any, comment_type: Any) -> None:
        if not hasattr(self, '_comment'):
            self._comment = [None, None, None]
        else:
            assert len(self._comment) == 3
            assert self._comment[1] is None
        if self.comment[1] is None:
            self._comment[1] = []  # type: ignore
        self._comment[1].extend([None] * (comment_type + 1 - len(self.comment[1])))  # type: ignore # NOQA
        # nprintf('commy', self.comment, comment_type)
        self._comment[1][comment_type] = comment  # type: ignore

    def add_comment_post(self, comment: Any) -> None:
        if not hasattr(self, '_comment'):
            self._comment = [None, None, []]  # type: ignore
        else:
            assert len(self._comment) == 3
            if self._comment[2] is None:
                self._comment[2] = []  # type: ignore
        self._comment[2].append(comment)  # type: ignore

    # def get_comment(self) -> Any:
    #     return getattr(self, '_comment', None)

    @property
    def comment(self) -> Any:
        return getattr(self, '_comment', None)

    def move_old_comment(self, target: Any, empty: bool = False) -> Any:
        """move a comment from this token to target (normally next token)
        used to combine e.g. comments before a BlockEntryToken to the
        ScalarToken that follows it
        empty is a special for empty values -> comment after key
        """
        c = self.comment
        if c is None:
            return
        # don't push beyond last element
        if isinstance(target, (StreamEndToken, DocumentStartToken)):
            return
        delattr(self, '_comment')
        tc = target.comment
        if not tc:  # target comment, just insert
            # special for empty value in key: value issue 25
            if empty:
                c = [c[0], c[1], None, None, c[0]]
            target._comment = c
            # nprint('mco2:', self, target, target.comment, empty)
            return self
        if c[0] and tc[0] or c[1] and tc[1]:
            if isinstance(c[1], list) and isinstance(tc[1], list):
                c[1].extend(tc[1])
            else:
                raise NotImplementedError(f'overlap in comment {c!r} {tc!r}')
        if c[0]:
            tc[0] = c[0]
        if c[1]:
            tc[1] = c[1]
        return self

    def split_old_comment(self) -> Any:
        """ split the post part of a comment, and return it
        as comment to be added. Delete second part if [None, None]
         abc:  # this goes to sequence
           # this goes to first element
           - first element
        """
        comment = self.comment
        if comment is None or comment[0] is None:
            return None  # nothing to do
        ret_val = [comment[0], None]
        if comment[1] is None:
            delattr(self, '_comment')
        return ret_val

    def move_new_comment(self, target: Any, empty: bool = False) -> Any:
        """move a comment from this token to target (normally next token)
        used to combine e.g. comments before a BlockEntryToken to the
        ScalarToken that follows it
        empty is a special for empty values -> comment after key
        """
        c = self.comment
        if c is None:
            return
        # don't push beyond last element
        if isinstance(target, (StreamEndToken, DocumentStartToken)):
            return
        delattr(self, '_comment')
        tc = target.comment
        if not tc:  # target comment, just insert
            # special for empty value in key: value issue 25
            if empty:
                c = [c[0], c[1], c[2]]
            target._comment = c
            # nprint('mco2:', self, target, target.comment, empty)
            return self
        # if self and target have both pre, eol or post comments, something seems wrong
        for idx in range(3):
            if c[idx] is not None and tc[idx] is not None:
                raise NotImplementedError(f'overlap in comment {c!r} {tc!r}')
        # move the comment parts
        for idx in range(3):
            if c[idx]:
                tc[idx] = c[idx]
        return self


# class BOMToken(Token):
#     id = '<byte order mark>'


class DirectiveToken(Token):
    __slots__ = 'name', 'value'
    id = '<directive>'

    def __init__(self, name: Any, value: Any, start_mark: Any, end_mark: Any) -> None:
        Token.__init__(self, start_mark, end_mark)
        self.name = name
        self.value = value


class DocumentStartToken(Token):
    __slots__ = ()
    id = '<document start>'


class DocumentEndToken(Token):
    __slots__ = ()
    id = '<document end>'


class StreamStartToken(Token):
    __slots__ = ('encoding',)
    id = '<stream start>'

    def __init__(
        self, start_mark: Any = None, end_mark: Any = None, encoding: Any = None,
    ) -> None:
        Token.__init__(self, start_mark, end_mark)
        self.encoding = encoding


class StreamEndToken(Token):
    __slots__ = ()
    id = '<stream end>'


class BlockSequenceStartToken(Token):
    __slots__ = ()
    id = '<block sequence start>'


class BlockMappingStartToken(Token):
    __slots__ = ()
    id = '<block mapping start>'


class BlockEndToken(Token):
    __slots__ = ()
    id = '<block end>'


class FlowSequenceStartToken(Token):
    __slots__ = ()
    id = '['


class FlowMappingStartToken(Token):
    __slots__ = ()
    id = '{'


class FlowSequenceEndToken(Token):
    __slots__ = ()
    id = ']'


class FlowMappingEndToken(Token):
    __slots__ = ()
    id = '}'


class KeyToken(Token):
    __slots__ = ()
    id = '?'

#   def x__repr__(self):
#       return f'KeyToken({self.start_mark.buffer[self.start_mark.index:].split(None, 1)[0]})'


class ValueToken(Token):
    __slots__ = ()
    id = ':'


class BlockEntryToken(Token):
    __slots__ = ()
    id = '-'


class FlowEntryToken(Token):
    __slots__ = ()
    id = ','


class AliasToken(Token):
    __slots__ = ('value',)
    id = '<alias>'

    def __init__(self, value: Any, start_mark: Any, end_mark: Any) -> None:
        Token.__init__(self, start_mark, end_mark)
        self.value = value


class AnchorToken(Token):
    __slots__ = ('value',)
    id = '<anchor>'

    def __init__(self, value: Any, start_mark: Any, end_mark: Any) -> None:
        Token.__init__(self, start_mark, end_mark)
        self.value = value


class TagToken(Token):
    __slots__ = ('value',)
    id = '<tag>'

    def __init__(self, value: Any, start_mark: Any, end_mark: Any) -> None:
        Token.__init__(self, start_mark, end_mark)
        self.value = value


class ScalarToken(Token):
    __slots__ = 'value', 'plain', 'style'
    id = '<scalar>'

    def __init__(
        self, value: Any, plain: Any, start_mark: Any, end_mark: Any, style: Any = None,
    ) -> None:
        Token.__init__(self, start_mark, end_mark)
        self.value = value
        self.plain = plain
        self.style = style


class CommentToken(Token):
    __slots__ = '_value', '_column', 'pre_done'
    id = '<comment>'

    def __init__(
        self, value: Any, start_mark: Any = None, end_mark: Any = None, column: Any = None,
    ) -> None:
        if start_mark is None:
            assert column is not None
            self._column = column
        Token.__init__(self, start_mark, end_mark)
        self._value = value

    @property
    def value(self) -> str:
        if isinstance(self._value, str):
            return self._value
        return "".join(self._value)

    @value.setter
    def value(self, val: Any) -> None:
        self._value = val

    def reset(self) -> None:
        if hasattr(self, 'pre_done'):
            delattr(self, 'pre_done')

    def __repr__(self) -> Any:
        v = f'{self.value!r}'
        if SHOW_LINES:
            try:
                v += ', line: ' + str(self.start_mark.line)
            except:  # NOQA
                pass
            try:
                v += ', col: ' + str(self.start_mark.column)
            except:  # NOQA
                pass
        return f'CommentToken({v})'

    def __eq__(self, other: Any) -> bool:
        if self.start_mark != other.start_mark:
            return False
        if self.end_mark != other.end_mark:
            return False
        if self.value != other.value:
            return False
        return True

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)



from pydantic import BaseModel, StrictStr, Field


class Config(BaseModel):
    attributes: dict[StrictStr, StrictStr] | None = None
    availability_zones: list[StrictStr] | None = None
    import_values: dict[StrictStr, StrictStr] | None = None
    mappings: dict[StrictStr, StrictStr] | None = None
    parameters: dict[StrictStr, StrictStr] | None = None
    references: dict[StrictStr, StrictStr] | None = None




from __future__ import annotations

"""
In round-trip mode the original tag needs to be preserved, but the tag
transformed based on the directives needs to be available as well.

A Tag that is created during loading has a handle and a suffix.
Not all objects loaded currently have a Tag, that .tag attribute can be None
A Tag that is created for dumping only (on an object loaded without a tag) has a suffix
only.
"""
from typing import Any, Dict, Optional, List, Union, Iterator  # NOQA

tag_attrib = '_yaml_tag'


class Tag:
    """store original tag information for roundtripping"""

    attrib = tag_attrib

    def __init__(self, handle: Any = None, suffix: Any = None, handles: Any = None) -> None:
        self.handle = handle
        self.suffix = suffix
        self.handles = handles
        self._transform_type: Optional[bool] = None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.trval!r})'

    def __str__(self) -> str:
        return f'{self.trval}'

    def __hash__(self) -> int:
        try:
            return self._hash_id  # type: ignore
        except AttributeError:
            self._hash_id = res = hash((self.handle, self.suffix))
            return res

    def __eq__(self, other: Any) -> bool:
        # other should not be a string, but the serializer sometimes provides these
        if isinstance(other, str):
            return self.trval == other
        return bool(self.trval == other.trval)

    def startswith(self, x: str) -> bool:
        if self.trval is not None:
            return self.trval.startswith(x)
        return False

    @property
    def trval(self) -> Optional[str]:
        try:
            return self._trval
        except AttributeError:
            pass
        if self.handle is None:
            self._trval: Optional[str] = self.uri_decoded_suffix
            return self._trval
        assert self._transform_type is not None
        if not self._transform_type:
            # the non-round-trip case
            self._trval = self.handles[self.handle] + self.uri_decoded_suffix
            return self._trval
        # round-trip case
        if self.handle == '!!' and self.suffix in (
            'null',
            'bool',
            'int',
            'float',
            'binary',
            'timestamp',
            'omap',
            'pairs',
            'set',
            'str',
            'seq',
            'map',
        ):
            self._trval = self.handles[self.handle] + self.uri_decoded_suffix
        else:
            # self._trval = self.handle + self.suffix
            self._trval = self.handles[self.handle] + self.uri_decoded_suffix
        return self._trval

    value = trval

    @property
    def uri_decoded_suffix(self) -> Optional[str]:
        try:
            return self._uri_decoded_suffix
        except AttributeError:
            pass
        if self.suffix is None:
            self._uri_decoded_suffix: Optional[str] = None
            return None
        res = ''
        # don't have to check for scanner errors here
        idx = 0
        while idx < len(self.suffix):
            ch = self.suffix[idx]
            idx += 1
            if ch != '%':
                res += ch
            else:
                res += chr(int(self.suffix[idx : idx + 2], 16))
                idx += 2
        self._uri_decoded_suffix = res
        return res

    def select_transform(self, val: bool) -> None:
        """
        val: False -> non-round-trip
             True -> round-trip
        """
        assert self._transform_type is None
        self._transform_type = val

    def check_handle(self) -> bool:
        if self.handle is None:
            return False
        return self.handle not in self.handles




from __future__ import annotations

import warnings

from cfn_check.yaml.util import configobj_walker as new_configobj_walker
from typing import Any


def configobj_walker(cfg: Any) -> Any:
    warnings.warn(
        'configobj_walker has moved to ruamel.yaml.util, please update your code',
        stacklevel=2,
    )
    return new_configobj_walker(cfg)



from collections import deque
from typing import Any
from cfn_check.yaml.comments import CommentedMap, CommentedSeq

from cfn_check.shared.types import (
    Data,
    Items,
    YamlObject,
)

from cfn_check.rendering import Renderer
from .parsing import QueryParser
from .parsing.token import Token

class Evaluator:

    def __init__(
        self,
        flags: list[str] | None = None
    ):
        if flags is None:
            flags = []

        self.flags = flags
        self._query_parser = QueryParser()
        self._renderer = Renderer()

    def match(
        self,
        resources: YamlObject,
        path: str,
        attributes: dict[str, Any] | None = None,
        availability_zones: list[str] | None = None,
        import_values: dict[str, tuple[str, CommentedMap]] | None = None,
        mappings: dict[str, str] | None = None,
        parameters: dict[str, Any] | None = None,
        references: dict[str, str] | None = None,
    ):
        items: Items = deque()
        
        if 'no-render' not in self.flags:
            resources = self._renderer.render(
                resources,
                attributes=attributes,
                availability_zones=availability_zones,
                import_values=import_values,
                mappings=mappings,
                parameters=parameters,
                references=references,
            )

        items.append(resources)

        segments = []
        for segment in path.split("."):
            segments.extend(self._query_parser.parse(segment))

        # Queries can be multi-segment,
        # so we effectively perform per-segment
        # repeated DFS searches, returning the matches
        # for each segment

        return self._search_document(resources, segments)

    def _search_document(
        self,
        root: Any,
        steps: list[Token],
    ) -> list[Any]:
        """
        Perform breadth-first search on a ruamel.yaml tree using a list of steps.
        
        At each level, only keeps nodes that have children matching the current step:
        - For CommentedMap: checks if any key matches the current step
        - For CommentedSeq: checks if any index matches the current step
        
        Args:
            root: Root of the ruamel.yaml tree
            steps: List of strings or integers representing steps to match
        
        Returns:
            List of nodes that match the full path of steps (all nodes at the final level)
        """
        if not steps:
            return [root]
        
        # Queue for BFS: (node, current_step_index)
        queue = deque([(root, 0)])
        matching_nodes = []
        path: list[str] = []

        while queue:
            node, step_idx = queue.popleft()
            
            # If we've consumed all steps, this node is a match
            if step_idx >= len(steps):
                matching_nodes.append(node)
                continue
            
            current_step = steps[step_idx]
            
            # Check if this node has children matching the current step
            if isinstance(node, (CommentedMap, dict)):
                (keys, found) = current_step.match(node)
                # Check all keys at this level
                if keys is not None and found is not None :
                    # This node has a matching child, add child to queue for next level
                    for found_key, found_val in zip(keys, found):
                        path.append(found_key)
                        queue.append((found_val, step_idx + 1))

            elif isinstance(node, (CommentedSeq, list)):
                (keys, found) = current_step.match(node)
                if keys is not None and found is not None :
                    for found_key in keys:
                        path.append(found_key)

                    for found_val in found:
                            queue.append((found_val, step_idx + 1))

        results: list[tuple[str, Data]] = []
        for idx, item in enumerate(list(matching_nodes)):
            results.append((
                path[idx],
                item,
            ))

        return results


from typing import Callable
from pydantic import ValidationError

from cfn_check.shared.types import Data
from cfn_check.evaluation.evaluator import Evaluator

class Collection:
    
    def __init__(self):
        self.documents: dict[str, Data] = {}
        self._evaluator = Evaluator()

    def query(
        self,
        query: str,
        document: str | None = None,
        transforms: list[Callable[[Data], Data]] | None = None
    ) -> list[Data] | None:

        if document and (
            document_data := self.documents.get(document)
        ):
            return self._evaluator.match(
                document_data,
                query,
            )
        
        results: list[tuple[str, Data]] = []

        for document_data in self.documents.values():
            result = self._evaluator.match(
                document_data,
                query,
            )

            results.extend(result)
        
        transformed: list[Data] = []
        if transforms:
            try:
                for _, found in results:
                    for transform in transforms:
                        found = transform(found)

                        if found is None:
                            return
                        
                    if found:
                        transformed.append(found)
            
                return transformed

            except ValidationError:
                pass

        return [
            found for _, found in results
        ]




from typing import TypeVar, Generic
from pydantic import BaseModel, ValidationError, JsonValue
from typing import Callable, get_type_hints

from cfn_check.shared.types import Data


T = TypeVar("T", bound= JsonValue | BaseModel)


class Validator(Generic[T]):
    def __init__(
        self,
        func: Callable[[T], None],
        query: str,
        name: str,
        transforms: list[Callable[[Data], Data]] | None = None
    ):
        self.func = func
        self.query = query
        self.name = name
        self.transforms = transforms

        self.model: BaseModel | None = None

        for _, param in get_type_hints(
            self.func,
        ).items():
            if param in BaseModel.__subclasses__():
                self.model = param

    def __call__(self, arg: tuple[str, Data]):

        try:
            path, item = arg

            if self.transforms:
                for transform in self.transforms:
                    item = transform(item)
                    
                    if item is None:
                        return

            if self.model and isinstance(item, dict):
                item = self.model(**item)

            return self.func(item)
        
        except ValidationError as err:
            return Exception(
                f'Path: {path}\n❌ Validation Error: {str(err)}'
            )
        
        except Exception as err:
            return Exception(
                f'Path: {path}\nError: {str(err)}'
            )




from __future__ import annotations

# The following YAML grammar is LL(1) and is parsed by a recursive descent
# parser.
#
# stream            ::= STREAM-START implicit_document? explicit_document*
#                                                                   STREAM-END
# implicit_document ::= block_node DOCUMENT-END*
# explicit_document ::= DIRECTIVE* DOCUMENT-START block_node? DOCUMENT-END*
# block_node_or_indentless_sequence ::=
#                       ALIAS
#                       | properties (block_content |
#                                                   indentless_block_sequence)?
#                       | block_content
#                       | indentless_block_sequence
# block_node        ::= ALIAS
#                       | properties block_content?
#                       | block_content
# flow_node         ::= ALIAS
#                       | properties flow_content?
#                       | flow_content
# properties        ::= TAG ANCHOR? | ANCHOR TAG?
# block_content     ::= block_collection | flow_collection | SCALAR
# flow_content      ::= flow_collection | SCALAR
# block_collection  ::= block_sequence | block_mapping
# flow_collection   ::= flow_sequence | flow_mapping
# block_sequence    ::= BLOCK-SEQUENCE-START (BLOCK-ENTRY block_node?)*
#                                                                   BLOCK-END
# indentless_sequence   ::= (BLOCK-ENTRY block_node?)+
# block_mapping     ::= BLOCK-MAPPING_START
#                       ((KEY block_node_or_indentless_sequence?)?
#                       (VALUE block_node_or_indentless_sequence?)?)*
#                       BLOCK-END
# flow_sequence     ::= FLOW-SEQUENCE-START
#                       (flow_sequence_entry FLOW-ENTRY)*
#                       flow_sequence_entry?
#                       FLOW-SEQUENCE-END
# flow_sequence_entry   ::= flow_node | KEY flow_node? (VALUE flow_node?)?
# flow_mapping      ::= FLOW-MAPPING-START
#                       (flow_mapping_entry FLOW-ENTRY)*
#                       flow_mapping_entry?
#                       FLOW-MAPPING-END
# flow_mapping_entry    ::= flow_node | KEY flow_node? (VALUE flow_node?)?
#
# FIRST sets:
#
# stream: { STREAM-START <}
# explicit_document: { DIRECTIVE DOCUMENT-START }
# implicit_document: FIRST(block_node)
# block_node: { ALIAS TAG ANCHOR SCALAR BLOCK-SEQUENCE-START
#                  BLOCK-MAPPING-START FLOW-SEQUENCE-START FLOW-MAPPING-START }
# flow_node: { ALIAS ANCHOR TAG SCALAR FLOW-SEQUENCE-START FLOW-MAPPING-START }
# block_content: { BLOCK-SEQUENCE-START BLOCK-MAPPING-START
#                               FLOW-SEQUENCE-START FLOW-MAPPING-START SCALAR }
# flow_content: { FLOW-SEQUENCE-START FLOW-MAPPING-START SCALAR }
# block_collection: { BLOCK-SEQUENCE-START BLOCK-MAPPING-START }
# flow_collection: { FLOW-SEQUENCE-START FLOW-MAPPING-START }
# block_sequence: { BLOCK-SEQUENCE-START }
# block_mapping: { BLOCK-MAPPING-START }
# block_node_or_indentless_sequence: { ALIAS ANCHOR TAG SCALAR
#               BLOCK-SEQUENCE-START BLOCK-MAPPING-START FLOW-SEQUENCE-START
#               FLOW-MAPPING-START BLOCK-ENTRY }
# indentless_sequence: { ENTRY }
# flow_collection: { FLOW-SEQUENCE-START FLOW-MAPPING-START }
# flow_sequence: { FLOW-SEQUENCE-START }
# flow_mapping: { FLOW-MAPPING-START }
# flow_sequence_entry: { ALIAS ANCHOR TAG SCALAR FLOW-SEQUENCE-START
#                                                    FLOW-MAPPING-START KEY }
# flow_mapping_entry: { ALIAS ANCHOR TAG SCALAR FLOW-SEQUENCE-START
#                                                    FLOW-MAPPING-START KEY }

# need to have full path with import, as pkg_resources tries to load parser.py in __init__.py
# only to not do anything with the package afterwards
# and for Jython too


from cfn_check.yaml.error import MarkedYAMLError
from cfn_check.yaml.tokens import (
    DocumentEndToken,
    DirectiveToken,
    DocumentStartToken,
    StreamEndToken,
    AliasToken,
    AnchorToken,
    TagToken,
    BlockEndToken,
    BlockEntryToken,
    BlockMappingStartToken,
    BlockSequenceStartToken,
    FlowEntryToken,
    FlowMappingEndToken,
    FlowSequenceEndToken,
    FlowMappingStartToken,
    FlowSequenceStartToken,
    KeyToken,
    ScalarToken,
    ValueToken,

)  # NOQA
from cfn_check.yaml.events import (
    DocumentStartEvent,
    SequenceStartEvent,
    MappingStartEvent,  
    DocumentEndEvent,
    SequenceEndEvent,
    StreamEndEvent,
    StreamStartEvent,
    AliasEvent,
    ScalarEvent,
    MappingEndEvent,
)

from cfn_check.yaml.scanner import Scanner, RoundTripScanner, ScannerError  # NOQA
from cfn_check.yaml.scanner import BlankLineComment
from cfn_check.yaml.comments import C_PRE, C_POST, C_SPLIT_ON_FIRST_BLANK
from cfn_check.yaml.compat import nprint, nprintf  # NOQA
from cfn_check.yaml.tag import Tag

from typing import Any, Dict, Optional, List, Optional  # NOQA

__all__ = ['Parser', 'RoundTripParser', 'ParserError']


def xprintf(*args: Any, **kw: Any) -> Any:
    return nprintf(*args, **kw)
    pass


class ParserError(MarkedYAMLError):
    pass


class Parser:
    # Since writing a recursive-descendant parser is a straightforward task, we
    # do not give many comments here.

    DEFAULT_TAGS = {'!': '!', '!!': 'tag:yaml.org,2002:'}

    def __init__(self, loader: Any) -> None:
        self.loader = loader
        if self.loader is not None and getattr(self.loader, '_parser', None) is None:
            self.loader._parser = self
        self.reset_parser()

    def reset_parser(self) -> None:
        # Reset the state attributes (to clear self-references)
        self.current_event = self.last_event = None
        self.tag_handles: Dict[Any, Any] = {}
        self.states: List[Any] = []
        self.marks: List[Any] = []
        self.state: Any = self.parse_stream_start

    def dispose(self) -> None:
        self.reset_parser()

    @property
    def scanner(self) -> Any:
        if hasattr(self.loader, 'typ'):
            return self.loader.scanner
        return self.loader._scanner

    @property
    def resolver(self) -> Any:
        if hasattr(self.loader, 'typ'):
            return self.loader.resolver
        return self.loader._resolver

    def check_event(self, *choices: Any) -> bool:
        # Check the type of the next event.
        if self.current_event is None:
            if self.state:
                self.current_event = self.state()
        if self.current_event is not None:
            if not choices:
                return True
            for choice in choices:
                if isinstance(self.current_event, choice):
                    return True
        return False

    def peek_event(self) -> Any:
        # Get the next event.
        if self.current_event is None:
            if self.state:
                self.current_event = self.state()
        return self.current_event

    def get_event(self) -> Any:
        # Get the next event and proceed further.
        if self.current_event is None:
            if self.state:
                self.current_event = self.state()
        # assert self.current_event is not None
        # if self.current_event.end_mark.line != self.peek_event().start_mark.line:
        # xprintf('get_event', repr(self.current_event), self.peek_event().start_mark.line)
        self.last_event = value = self.current_event
        self.current_event = None
        return value

    # stream    ::= STREAM-START implicit_document? explicit_document*
    #                                                               STREAM-END
    # implicit_document ::= block_node DOCUMENT-END*
    # explicit_document ::= DIRECTIVE* DOCUMENT-START block_node? DOCUMENT-END*

    def parse_stream_start(self) -> Any:
        # Parse the stream start.
        token = self.scanner.get_token()
        self.move_token_comment(token)
        event = StreamStartEvent(token.start_mark, token.end_mark, encoding=token.encoding)

        # Prepare the next state.
        self.state = self.parse_implicit_document_start

        return event

    def parse_implicit_document_start(self) -> Any:
        # Parse an implicit document.
        if not self.scanner.check_token(DirectiveToken, DocumentStartToken, StreamEndToken):
            # don't need copy, as an implicit tag doesn't add tag_handles
            self.tag_handles = self.DEFAULT_TAGS
            token = self.scanner.peek_token()
            start_mark = end_mark = token.start_mark
            event = DocumentStartEvent(start_mark, end_mark, explicit=False)

            # Prepare the next state.
            self.states.append(self.parse_document_end)
            self.state = self.parse_block_node

            return event

        else:
            return self.parse_document_start()

    def parse_document_start(self) -> Any:
        # Parse any extra document end indicators.
        while self.scanner.check_token(DocumentEndToken):
            self.scanner.get_token()
        # Parse an explicit document.
        if not self.scanner.check_token(StreamEndToken):
            version, tags = self.process_directives()
            if not self.scanner.check_token(DocumentStartToken):
                raise ParserError(
                    None,
                    None,
                    "expected '<document start>', "
                    f'but found {self.scanner.peek_token().id,!r}',
                    self.scanner.peek_token().start_mark,
                )
            token = self.scanner.get_token()
            start_mark = token.start_mark
            end_mark = token.end_mark
            # if self.loader is not None and \
            #    end_mark.line != self.scanner.peek_token().start_mark.line:
            #     self.loader.scalar_after_indicator = False
            event: Any = DocumentStartEvent(
                start_mark,
                end_mark,
                explicit=True,
                version=version,
                tags=tags,
                comment=token.comment,
            )
            self.states.append(self.parse_document_end)
            self.state = self.parse_document_content
        else:
            # Parse the end of the stream.
            token = self.scanner.get_token()
            event = StreamEndEvent(token.start_mark, token.end_mark, comment=token.comment)
            assert not self.states
            assert not self.marks
            self.state = None
        return event

    def parse_document_end(self) -> Any:
        # Parse the document end.
        token = self.scanner.peek_token()
        start_mark = end_mark = token.start_mark
        explicit = False
        if self.scanner.check_token(DocumentEndToken):
            token = self.scanner.get_token()
            # if token.end_mark.line != self.peek_event().start_mark.line:
            pt = self.scanner.peek_token()
            if not isinstance(pt, StreamEndToken) and (
                token.end_mark.line == pt.start_mark.line
            ):
                raise ParserError(
                    None,
                    None,
                    'found non-comment content after document end marker, '
                    f'{self.scanner.peek_token().id,!r}',
                    self.scanner.peek_token().start_mark,
                )
            end_mark = token.end_mark
            explicit = True
        event = DocumentEndEvent(start_mark, end_mark, explicit=explicit)

        # Prepare the next state.
        if self.resolver.processing_version == (1, 1):
            self.state = self.parse_document_start
        else:
            if explicit:
                # found a document end marker, can be followed by implicit document
                self.state = self.parse_implicit_document_start
            else:
                self.state = self.parse_document_start

        return event

    def parse_document_content(self) -> Any:
        if self.scanner.check_token(
            DirectiveToken, DocumentStartToken, DocumentEndToken, StreamEndToken,
        ):
            event = self.process_empty_scalar(self.scanner.peek_token().start_mark)
            self.state = self.states.pop()
            return event
        else:
            return self.parse_block_node()

    def process_directives(self) -> Any:
        yaml_version = None
        self.tag_handles = {}
        while self.scanner.check_token(DirectiveToken):
            token = self.scanner.get_token()
            if token.name == 'YAML':
                if yaml_version is not None:
                    raise ParserError(
                        None, None, 'found duplicate YAML directive', token.start_mark,
                    )
                major, minor = token.value
                if major != 1:
                    raise ParserError(
                        None,
                        None,
                        'found incompatible YAML document (version 1.* is required)',
                        token.start_mark,
                    )
                yaml_version = token.value
            elif token.name == 'TAG':
                handle, prefix = token.value
                if handle in self.tag_handles:
                    raise ParserError(
                        None, None, f'duplicate tag handle {handle!r}', token.start_mark,
                    )
                self.tag_handles[handle] = prefix
        if bool(self.tag_handles):
            value: Any = (yaml_version, self.tag_handles.copy())
        else:
            value = yaml_version, None
        if self.loader is not None and hasattr(self.loader, 'tags'):
            # ToDo: this is used to keep a single loaded file from losing its version
            # info, but  it affects following versions that have no explicit directive
            self.loader.version = yaml_version
            if self.loader.tags is None:
                self.loader.tags = {}
            for k in self.tag_handles:
                self.loader.tags[k] = self.tag_handles[k]
                self.loader.doc_infos[-1].tags.append((k, self.tag_handles[k]))
        for key in self.DEFAULT_TAGS:
            if key not in self.tag_handles:
                self.tag_handles[key] = self.DEFAULT_TAGS[key]
        return value

    # block_node_or_indentless_sequence ::= ALIAS
    #               | properties (block_content | indentless_block_sequence)?
    #               | block_content
    #               | indentless_block_sequence
    # block_node    ::= ALIAS
    #                   | properties block_content?
    #                   | block_content
    # flow_node     ::= ALIAS
    #                   | properties flow_content?
    #                   | flow_content
    # properties    ::= TAG ANCHOR? | ANCHOR TAG?
    # block_content     ::= block_collection | flow_collection | SCALAR
    # flow_content      ::= flow_collection | SCALAR
    # block_collection  ::= block_sequence | block_mapping
    # flow_collection   ::= flow_sequence | flow_mapping

    def parse_block_node(self) -> Any:
        return self.parse_node(block=True)

    def parse_flow_node(self) -> Any:
        return self.parse_node()

    def parse_block_node_or_indentless_sequence(self) -> Any:
        return self.parse_node(block=True, indentless_sequence=True)

    # def transform_tag(self, handle: Any, suffix: Any) -> Any:
    #     return self.tag_handles[handle] + suffix

    def select_tag_transform(self, tag: Tag) -> None:
        if tag is None:
            return
        tag.select_transform(False)

    def parse_node(self, block: bool = False, indentless_sequence: bool = False) -> Any:
        if self.scanner.check_token(AliasToken):
            token = self.scanner.get_token()
            event: Any = AliasEvent(token.value, token.start_mark, token.end_mark)
            self.state = self.states.pop()
            return event

        anchor = None
        tag = None
        start_mark = end_mark = tag_mark = None
        if self.scanner.check_token(AnchorToken):
            token = self.scanner.get_token()
            self.move_token_comment(token)
            start_mark = token.start_mark
            end_mark = token.end_mark
            anchor = token.value
            if self.scanner.check_token(TagToken):
                token = self.scanner.get_token()
                tag_mark = token.start_mark
                end_mark = token.end_mark
                # tag = token.value
                tag = Tag(
                    handle=token.value[0], suffix=token.value[1], handles=self.tag_handles,
                )
        elif self.scanner.check_token(TagToken):
            token = self.scanner.get_token()
            try:
                self.move_token_comment(token)
            except NotImplementedError:
                pass
            start_mark = tag_mark = token.start_mark
            end_mark = token.end_mark
            # tag = token.value
            tag = Tag(handle=token.value[0], suffix=token.value[1], handles=self.tag_handles)
            if self.scanner.check_token(AnchorToken):
                token = self.scanner.get_token()
                start_mark = tag_mark = token.start_mark
                end_mark = token.end_mark
                anchor = token.value
        if tag is not None:
            self.select_tag_transform(tag)
            if tag.check_handle():
                raise ParserError(
                    'while parsing a node',
                    start_mark,
                    f'found undefined tag handle {tag.handle!r}',
                    tag_mark,
                )
        if start_mark is None:
            start_mark = end_mark = self.scanner.peek_token().start_mark
        event = None
        implicit = tag is None or str(tag) == '!'
        if indentless_sequence and self.scanner.check_token(BlockEntryToken):
            comment = None
            pt = self.scanner.peek_token()
            if self.loader and self.loader.comment_handling is None:
                if pt.comment and pt.comment[0]:
                    comment = [pt.comment[0], []]
                    pt.comment[0] = None
                elif pt.comment and pt.comment[0] is None and pt.comment[1]:
                    comment = [None, pt.comment[1]]
                    pt.comment[1] = None
            elif self.loader:
                if pt.comment:
                    comment = pt.comment
            end_mark = self.scanner.peek_token().end_mark
            event = SequenceStartEvent(
                anchor, tag, implicit, start_mark, end_mark, flow_style=False, comment=comment,
            )
            self.state = self.parse_indentless_sequence_entry
            return event

        if self.scanner.check_token(ScalarToken):
            token = self.scanner.get_token()
            # self.scanner.peek_token_same_line_comment(token)
            end_mark = token.end_mark
            if (token.plain and tag is None) or str(tag) == '!':
                dimplicit = (True, False)
            elif tag is None:
                dimplicit = (False, True)
            else:
                dimplicit = (False, False)
            event = ScalarEvent(
                anchor,
                tag,
                dimplicit,
                token.value,
                start_mark,
                end_mark,
                style=token.style,
                comment=token.comment,
            )
            self.state = self.states.pop()
        elif self.scanner.check_token(FlowSequenceStartToken):
            pt = self.scanner.peek_token()
            end_mark = pt.end_mark
            event = SequenceStartEvent(
                anchor,
                tag,
                implicit,
                start_mark,
                end_mark,
                flow_style=True,
                comment=pt.comment,
            )
            self.state = self.parse_flow_sequence_first_entry
        elif self.scanner.check_token(FlowMappingStartToken):
            pt = self.scanner.peek_token()
            end_mark = pt.end_mark
            event = MappingStartEvent(
                anchor,
                tag,
                implicit,
                start_mark,
                end_mark,
                flow_style=True,
                comment=pt.comment,
            )
            self.state = self.parse_flow_mapping_first_key
        elif block and self.scanner.check_token(BlockSequenceStartToken):
            end_mark = self.scanner.peek_token().start_mark
            # should inserting the comment be dependent on the
            # indentation?
            pt = self.scanner.peek_token()
            comment = pt.comment
            # nprint('pt0', type(pt))
            if comment is None or comment[1] is None:
                comment = pt.split_old_comment()
            # nprint('pt1', comment)
            event = SequenceStartEvent(
                anchor, tag, implicit, start_mark, end_mark, flow_style=False, comment=comment,
            )
            self.state = self.parse_block_sequence_first_entry
        elif block and self.scanner.check_token(BlockMappingStartToken):
            end_mark = self.scanner.peek_token().start_mark
            comment = self.scanner.peek_token().comment
            event = MappingStartEvent(
                anchor, tag, implicit, start_mark, end_mark, flow_style=False, comment=comment,
            )
            self.state = self.parse_block_mapping_first_key
        elif anchor is not None or tag is not None:
            # Empty scalars are allowed even if a tag or an anchor is
            # specified.
            event = ScalarEvent(anchor, tag, (implicit, False), "", start_mark, end_mark)
            self.state = self.states.pop()
        else:
            if block:
                node = 'block'
            else:
                node = 'flow'
            token = self.scanner.peek_token()
            raise ParserError(
                f'while parsing a {node!s} node',
                start_mark,
                f'expected the node content, but found {token.id!r}',
                token.start_mark,
            )
        return event

    # block_sequence ::= BLOCK-SEQUENCE-START (BLOCK-ENTRY block_node?)*
    #                                                               BLOCK-END

    def parse_block_sequence_first_entry(self) -> Any:
        token = self.scanner.get_token()
        # move any comment from start token
        # self.move_token_comment(token)
        self.marks.append(token.start_mark)
        return self.parse_block_sequence_entry()

    def parse_block_sequence_entry(self) -> Any:
        if self.scanner.check_token(BlockEntryToken):
            token = self.scanner.get_token()
            self.move_token_comment(token)
            if not self.scanner.check_token(BlockEntryToken, BlockEndToken):
                self.states.append(self.parse_block_sequence_entry)
                return self.parse_block_node()
            else:
                self.state = self.parse_block_sequence_entry
                return self.process_empty_scalar(token.end_mark)
        if not self.scanner.check_token(BlockEndToken):
            token = self.scanner.peek_token()
            raise ParserError(
                'while parsing a block collection',
                self.marks[-1],
                f'expected <block end>, but found {token.id!r}',
                token.start_mark,
            )
        token = self.scanner.get_token()  # BlockEndToken
        event = SequenceEndEvent(token.start_mark, token.end_mark, comment=token.comment)
        self.state = self.states.pop()
        self.marks.pop()
        return event

    # indentless_sequence ::= (BLOCK-ENTRY block_node?)+

    # indentless_sequence?
    # sequence:
    # - entry
    #  - nested

    def parse_indentless_sequence_entry(self) -> Any:
        if self.scanner.check_token(BlockEntryToken):
            token = self.scanner.get_token()
            self.move_token_comment(token)
            if not self.scanner.check_token(
                BlockEntryToken, KeyToken, ValueToken, BlockEndToken,
            ):
                self.states.append(self.parse_indentless_sequence_entry)
                return self.parse_block_node()
            else:
                self.state = self.parse_indentless_sequence_entry
                return self.process_empty_scalar(token.end_mark)
        token = self.scanner.peek_token()
        c = None
        if self.loader and self.loader.comment_handling is None:
            c = token.comment
            start_mark = token.start_mark
        else:
            start_mark = self.last_event.end_mark  # type: ignore
            c = self.distribute_comment(token.comment, start_mark.line)  # type: ignore
        event = SequenceEndEvent(start_mark, start_mark, comment=c)
        self.state = self.states.pop()
        return event

    # block_mapping     ::= BLOCK-MAPPING_START
    #                       ((KEY block_node_or_indentless_sequence?)?
    #                       (VALUE block_node_or_indentless_sequence?)?)*
    #                       BLOCK-END

    def parse_block_mapping_first_key(self) -> Any:
        token = self.scanner.get_token()
        self.marks.append(token.start_mark)
        return self.parse_block_mapping_key()

    def parse_block_mapping_key(self) -> Any:
        if self.scanner.check_token(KeyToken):
            token = self.scanner.get_token()
            self.move_token_comment(token)
            if not self.scanner.check_token(KeyToken, ValueToken, BlockEndToken):
                self.states.append(self.parse_block_mapping_value)
                return self.parse_block_node_or_indentless_sequence()
            else:
                self.state = self.parse_block_mapping_value
                return self.process_empty_scalar(token.end_mark)
        if self.resolver.processing_version > (1, 1) and self.scanner.check_token(ValueToken):
            self.state = self.parse_block_mapping_value
            return self.process_empty_scalar(self.scanner.peek_token().start_mark)
        
        while self.scanner.check_token(TagToken) and (
            token := self.scanner.get_token()
        ):
            if not self.scanner.check_token(KeyToken, ValueToken, BlockEndToken) and (
                token := self.scanner.peek_token()
            ):
                self.states.append(self.parse_block_mapping_value)
                return self.parse_block_node_or_indentless_sequence()
            else:
                token = self.scanner.peek_token()
                print('GOT AGAIN:', token)
                self.state = self.parse_block_mapping_value
                return self.process_empty_scalar(token.end_mark)
                
        if not self.scanner.check_token(BlockEndToken):
            token = self.scanner.peek_token()
            raise ParserError(
                'while parsing a block mapping',
                self.marks[-1],
                f'expected <block end>, but found {token.id!r}',
                token.start_mark,
            )
        token = self.scanner.get_token()
        self.move_token_comment(token)
        event = MappingEndEvent(token.start_mark, token.end_mark, comment=token.comment)
        self.state = self.states.pop()
        self.marks.pop()
        return event

    def parse_block_mapping_value(self) -> Any:
        if self.scanner.check_token(ValueToken):
            token = self.scanner.get_token()
            # value token might have post comment move it to e.g. block
            if self.scanner.check_token(ValueToken):
                self.move_token_comment(token)
            else:
                if not self.scanner.check_token(KeyToken):
                    self.move_token_comment(token, empty=True)
                # else: empty value for this key cannot move token.comment
            if not self.scanner.check_token(KeyToken, ValueToken, BlockEndToken):
                self.states.append(self.parse_block_mapping_key)
                return self.parse_block_node_or_indentless_sequence()
            else:
                self.state = self.parse_block_mapping_key
                comment = token.comment
                if comment is None:
                    token = self.scanner.peek_token()
                    comment = token.comment
                    if comment:
                        token._comment = [None, comment[1]]
                        comment = [comment[0], None]
                return self.process_empty_scalar(token.end_mark, comment=comment)
        else:
            self.state = self.parse_block_mapping_key
            token = self.scanner.peek_token()
            return self.process_empty_scalar(token.start_mark)

    # flow_sequence     ::= FLOW-SEQUENCE-START
    #                       (flow_sequence_entry FLOW-ENTRY)*
    #                       flow_sequence_entry?
    #                       FLOW-SEQUENCE-END
    # flow_sequence_entry   ::= flow_node | KEY flow_node? (VALUE flow_node?)?
    #
    # Note that while production rules for both flow_sequence_entry and
    # flow_mapping_entry are equal, their interpretations are different.
    # For `flow_sequence_entry`, the part `KEY flow_node? (VALUE flow_node?)?`
    # generate an inline mapping (set syntax).

    def parse_flow_sequence_first_entry(self) -> Any:
        token = self.scanner.get_token()
        self.marks.append(token.start_mark)
        return self.parse_flow_sequence_entry(first=True)

    def parse_flow_sequence_entry(self, first: bool = False) -> Any:
        if not self.scanner.check_token(FlowSequenceEndToken):
            if not first:
                if self.scanner.check_token(FlowEntryToken):
                    self.scanner.get_token()
                else:
                    token = self.scanner.peek_token()
                    raise ParserError(
                        'while parsing a flow sequence',
                        self.marks[-1],
                        f"expected ',' or ']', but got {token.id!r}",
                        token.start_mark,
                    )

            if self.scanner.check_token(KeyToken):
                token = self.scanner.peek_token()
                event: Any = MappingStartEvent(
                    None, None, True, token.start_mark, token.end_mark, flow_style=True,
                )
                self.state = self.parse_flow_sequence_entry_mapping_key
                return event
            elif not self.scanner.check_token(FlowSequenceEndToken):
                self.states.append(self.parse_flow_sequence_entry)
                return self.parse_flow_node()
        token = self.scanner.get_token()
        event = SequenceEndEvent(token.start_mark, token.end_mark, comment=token.comment)
        self.state = self.states.pop()
        self.marks.pop()
        return event

    def parse_flow_sequence_entry_mapping_key(self) -> Any:
        token = self.scanner.get_token()
        if not self.scanner.check_token(ValueToken, FlowEntryToken, FlowSequenceEndToken):
            self.states.append(self.parse_flow_sequence_entry_mapping_value)
            return self.parse_flow_node()
        else:
            self.state = self.parse_flow_sequence_entry_mapping_value
            return self.process_empty_scalar(token.end_mark)

    def parse_flow_sequence_entry_mapping_value(self) -> Any:
        if self.scanner.check_token(ValueToken):
            token = self.scanner.get_token()
            if not self.scanner.check_token(FlowEntryToken, FlowSequenceEndToken):
                self.states.append(self.parse_flow_sequence_entry_mapping_end)
                return self.parse_flow_node()
            else:
                self.state = self.parse_flow_sequence_entry_mapping_end
                return self.process_empty_scalar(token.end_mark)
        else:
            self.state = self.parse_flow_sequence_entry_mapping_end
            token = self.scanner.peek_token()
            return self.process_empty_scalar(token.start_mark)

    def parse_flow_sequence_entry_mapping_end(self) -> Any:
        self.state = self.parse_flow_sequence_entry
        token = self.scanner.peek_token()
        return MappingEndEvent(token.start_mark, token.start_mark)

    # flow_mapping  ::= FLOW-MAPPING-START
    #                   (flow_mapping_entry FLOW-ENTRY)*
    #                   flow_mapping_entry?
    #                   FLOW-MAPPING-END
    # flow_mapping_entry    ::= flow_node | KEY flow_node? (VALUE flow_node?)?

    def parse_flow_mapping_first_key(self) -> Any:
        token = self.scanner.get_token()
        self.marks.append(token.start_mark)
        return self.parse_flow_mapping_key(first=True)

    def parse_flow_mapping_key(self, first: Any = False) -> Any:
        if not self.scanner.check_token(FlowMappingEndToken):
            if not first:
                if self.scanner.check_token(FlowEntryToken):
                    self.scanner.get_token()
                else:
                    token = self.scanner.peek_token()
                    raise ParserError(
                        'while parsing a flow mapping',
                        self.marks[-1],
                        f"expected ',' or '}}', but got {token.id!r}",
                        token.start_mark,
                    )
            if self.scanner.check_token(KeyToken):
                token = self.scanner.get_token()
                if not self.scanner.check_token(
                    ValueToken, FlowEntryToken, FlowMappingEndToken,
                ):
                    self.states.append(self.parse_flow_mapping_value)
                    return self.parse_flow_node()
                else:
                    self.state = self.parse_flow_mapping_value
                    return self.process_empty_scalar(token.end_mark)
            elif self.resolver.processing_version > (1, 1) and self.scanner.check_token(
                ValueToken,
            ):
                self.state = self.parse_flow_mapping_value
                return self.process_empty_scalar(self.scanner.peek_token().end_mark)
            elif not self.scanner.check_token(FlowMappingEndToken):
                self.states.append(self.parse_flow_mapping_empty_value)
                return self.parse_flow_node()
        token = self.scanner.get_token()
        event = MappingEndEvent(token.start_mark, token.end_mark, comment=token.comment)
        self.state = self.states.pop()
        self.marks.pop()
        return event

    def parse_flow_mapping_value(self) -> Any:
        if self.scanner.check_token(ValueToken):
            token = self.scanner.get_token()
            if not self.scanner.check_token(FlowEntryToken, FlowMappingEndToken):
                self.states.append(self.parse_flow_mapping_key)
                return self.parse_flow_node()
            else:
                self.state = self.parse_flow_mapping_key
                return self.process_empty_scalar(token.end_mark)
        else:
            self.state = self.parse_flow_mapping_key
            token = self.scanner.peek_token()
            return self.process_empty_scalar(token.start_mark)

    def parse_flow_mapping_empty_value(self) -> Any:
        self.state = self.parse_flow_mapping_key
        return self.process_empty_scalar(self.scanner.peek_token().start_mark)

    def process_empty_scalar(self, mark: Any, comment: Any = None) -> Any:
        return ScalarEvent(None, None, (True, False), "", mark, mark, comment=comment)

    def move_token_comment(
        self, token: Any, nt: Optional[Any] = None, empty: Optional[bool] = False,
    ) -> Any:
        pass


class RoundTripParser(Parser):
    """roundtrip is a safe loader, that wants to see the unmangled tag"""

    def select_tag_transform(self, tag: Tag) -> None:
        if tag is None:
            return
        tag.select_transform(True)

    def move_token_comment(
        self, token: Any, nt: Optional[Any] = None, empty: Optional[bool] = False,
    ) -> Any:
        token.move_old_comment(self.scanner.peek_token() if nt is None else nt, empty=empty)


class RoundTripParserSC(RoundTripParser):
    """roundtrip is a safe loader, that wants to see the unmangled tag"""

    # some of the differences are based on the superclass testing
    # if self.loader.comment_handling is not None

    def move_token_comment(
        self: Any, token: Any, nt: Any = None, empty: Optional[bool] = False,
    ) -> None:
        token.move_new_comment(self.scanner.peek_token() if nt is None else nt, empty=empty)

    def distribute_comment(self, comment: Any, line: Any) -> Any:
        # ToDo, look at indentation of the comment to determine attachment
        if comment is None:
            return None
        if not comment[0]:
            return None
        # if comment[0][0] != line + 1:
        #     nprintf('>>>dcxxx', comment, line)
        assert comment[0][0] == line + 1
        # if comment[0] - line > 1:
        #     return
        typ = self.loader.comment_handling & 0b11
        # nprintf('>>>dca', comment, line, typ)
        if typ == C_POST:
            return None
        if typ == C_PRE:
            c = [None, None, comment[0]]
            comment[0] = None
            return c
        # nprintf('>>>dcb', comment[0])
        for _idx, cmntidx in enumerate(comment[0]):
            # nprintf('>>>dcb', cmntidx)
            if isinstance(self.scanner.comments[cmntidx], BlankLineComment):
                break
        else:
            return None  # no space found
        if _idx == 0:
            return None  # first line was blank
        # nprintf('>>>dcc', idx)
        if typ == C_SPLIT_ON_FIRST_BLANK:
            c = [None, None, comment[0][:_idx]]
            comment[0] = comment[0][_idx:]
            return c
        raise NotImplementedError  # reserved



class IPv4CIDRSolver:

    def __init__(
        self,
        host: str,
        desired: int,
        bits: int,
    ):
        self.host = host
        self.subnets_desired = desired
        self.subnet_bits = bits

        host_ip, mask = self.host.split('/', maxsplit=1)

        self.host_ip = host_ip
        self._host_mask_string = f'/{mask}'
        self.host_mask = int(mask)

        self.subnet_mask = 32 - bits

        self._host_octets = [
            int(octet) for octet in self.host.strip(self._host_mask_string).split('.')
        ]
    
    def provision_subnets(self):
        subnet_requested_ips = 2**self.subnet_bits
        host_available_ips = 2**(32 - self.host_mask)

        total_ips_requested = subnet_requested_ips * self.subnets_desired
        if host_available_ips < total_ips_requested:
            return []

        return [
            self._provision_subnet(
                subnet_requested_ips,
                idx
            ) for idx in range(self.subnets_desired)
        ]

    
    def _provision_subnet(
        self,
        requested_ips: int,
        idx: int,
    ):
        increment = requested_ips
        octet_idx = -1
        if requested_ips > 255:
            increment /= 256
            octet_idx -= 1

        increment *= idx

        subnet = list(self._host_octets)

        subnet[octet_idx] += increment

        subnet_base_ip = '.'.join([
            str(octet) for octet in subnet
        ])

        return f'{subnet_base_ip}/{self.subnet_mask}'



    



import inspect

from async_logging import LogLevelName, Logger, LoggingConfig
from cocoa.cli import CLI, ImportType, YamlFile

from cfn_check.cli.utils.attributes import bind
from cfn_check.cli.utils.files import load_templates, write_to_file
from cfn_check.evaluation.validate import ValidationSet
from cfn_check.logging.models import InfoLog
from cfn_check.collection.collection import Collection
from cfn_check.validation.validator import Validator
from .config import Config


@CLI.command(
    shortnames={
        'flags': 'F'
    },
    display_help_on_error=False
)
async def validate(
    path: str,
    config: YamlFile[Config] = 'config.yml',
    file_pattern: str | None = None,
    rules: ImportType[Collection] = None,
    flags: list[str] | None = None,
    log_level: LogLevelName = 'info',
):
    '''
    Validate Cloud Foundation
    
    @param config A CFN-Check yaml config file
    @param disabled A list of string features to disable during checks
    @param file_pattern A string pattern used to find template files
    @param rules Path to a file containing Collections
    @param log_level The log level to use
    '''

    config_data = config.data
    if config_data is None:
        config_data = Config()

    logging_config = LoggingConfig()
    logging_config.update(
        log_level=log_level,
        log_output='stderr',
    )

    logger = Logger()

    if flags is None:
        flags = []

    templates = await load_templates(
        path,
        file_pattern=file_pattern,
    )

    for file, data in templates:
        for name, rule in rules.data.items():
            rules.data[name] = rule()
            rules.data[name].documents[file] = data

    validation_set = ValidationSet([ 
        bind(
            rule,
            validation,
        )
        for rule in rules.data.values()
        for _, validation in inspect.getmembers(rule)
        if isinstance(validation, Validator)
    ], flags=flags)
    
    if validation_error := validation_set.validate([
        template_data for _, template_data in templates
    ]):
        raise validation_error
    
    templates_evaluated = len(templates)
    
    await logger.log(InfoLog(message=f'✅ {validation_set.count} validations met for {templates_evaluated} templates'))
    
    if config_data:
        await write_to_file(
            config.value,
            config_data.model_dump(),
        )



from __future__ import annotations

from typing import Optional, Tuple, Any

# from dataclasses import dataclass, field, MISSING  # NOQA


# @dataclass(order=True, frozen=True)
class Version:
    # major: int
    # minor: int
    def __init__(self, major: int, minor: int) -> None:
        self._major = major
        self._minor = minor

    @property
    def major(self) -> int:
        return self._major

    @property
    def minor(self) -> int:
        return self._minor

    def __eq__(self, v: Any) -> bool:
        if not isinstance(v, Version):
            return False
        return self._major == v._major and self._minor == v._minor

    def __lt__(self, v: Version) -> bool:
        if self._major < v._major:
            return True
        if self._major > v._major:
            return False
        return self._minor < v._minor

    def __le__(self, v: Version) -> bool:
        if self._major < v._major:
            return True
        if self._major > v._major:
            return False
        return self._minor <= v._minor

    def __gt__(self, v: Version) -> bool:
        if self._major > v._major:
            return True
        if self._major < v._major:
            return False
        return self._minor > v._minor

    def __ge__(self, v: Version) -> bool:
        if self._major > v._major:
            return True
        if self._major < v._major:
            return False
        return self._minor >= v._minor


def version(
    major: int | str | Tuple[int, int] | None,
    minor: Optional[int] = None,
) -> Optional[Version]:
    if major is None:
        assert minor is None
        return None
    if isinstance(major, str):
        assert minor is None
        parts = major.split('.')
        assert len(parts) == 2
        return Version(int(parts[0]), int(parts[1]))
    elif isinstance(major, tuple):
        assert minor is None
        assert len(major) == 2
        major, minor = major
    assert minor is not None
    return Version(major, minor)


# @dataclass(frozen=True)
class Tag:
    # handle: str
    # prefix: str
    def __init__(self, handle: str, prefix: str) -> None:
        self._handle = handle
        self._prefix = prefix

    @property
    def handle(self) -> str:
        return self._handle

    @property
    def prefix(self) -> str:
        return self._prefix


# @dataclass
class DocInfo:
    """
    Store document information, can be used for analysis of a loaded YAML document
    requested_version: if explicitly set before load
    doc_version: from %YAML directive
    tags: from %TAG directives in scanned order
    """
    # requested_version: Optional[Version] = None
    # doc_version: Optional[Version] = None
    # tags: list[Tag] = field(default_factory=list)
    def __init__(
        self,
        requested_version: Optional[Version] = None,
        doc_version: Optional[Version] = None,
        tags: Optional[list[Tag]] = None,
    ):
        self.requested_version = requested_version
        self.doc_version = doc_version
        self.tags = [] if tags is None else tags



from __future__ import annotations
import re
import sys
from collections import deque
from typing import Deque, Any
from cfn_check.shared.types import Data, Items
from cfn_check.yaml.comments import CommentedMap, CommentedSeq
from .token_type import TokenType
from .operators import ValueOperator

class Token:

    def __init__(
        self,
        selector: tuple[int, int] | int | re.Pattern | str,
        selector_type: TokenType,
        nested: list[Token] | None = None
    ):
        self.selector = selector
        self.selector_type = selector_type
        self._nested = nested
        self._block_or_chars = re.compile(r'\|')
        self._block_and_chars = re.compile(r'&')

        segments = self._block_or_chars.split(self.selector)

        self._selector_segments: list[list[str]] = [
            self._block_and_chars.split(segment)
            for segment in segments
        ]

    def match(
        self,
        node: Data,
    ):
        
        if self.selector_type in [
            TokenType.WILDCARD,
            TokenType.WILDCARD_RANGE,
        ] and isinstance(node, dict):
            return list(node.keys()), list(node.values())
        
        elif self.selector_type in [
            TokenType.WILDCARD,
            TokenType.WILDCARD_RANGE,
        ] and isinstance(node, list):
            return list(range(node)), node
        
        elif self.selector_type in [
            TokenType.WILDCARD,
            TokenType.WILDCARD_RANGE,
        ]:
            return None, node

        match self.selector_type:

            case TokenType.BOUND_RANGE:
                return self._match_bound_range(node)

            case TokenType.INDEX:
                return self._match_index(node)

            case TokenType.KEY:
                return self._match_key(node)
            
            case TokenType.KEY_RANGE:
                return self._match_key_range(node)
            
            case TokenType.NESTED_MAP:
                return self._match_nested_map(node)
            
            case TokenType.NESTED_RANGE:
                return self._match_nested_range(node)
        
            case TokenType.PATTERN:
                return self._match_pattern(node)

            case TokenType.PATTERN_RANGE:
                return self._match_pattern_range(node)

            case TokenType.UNBOUND_RANGE:
                return self._match_unbound_range(node)
            
            case TokenType.VALUE_MATCH:
                return self._match_value(node)

            case _:
                return None, None

    def _match_bound_range(
        self,
        node: Data,
    ):
        if not isinstance(node, CommentedSeq) or not isinstance(self.selector, tuple):
            return None, None
        
        start, stop = self.selector

        if stop == sys.maxsize:
            stop = len(node)

        return [f'{start}-{stop}'], list(node[start:stop])
    
    def _match_index(
        self,
        node: Data,
    ):
        if (
            isinstance(node, list)
        ) and (
            isinstance(self.selector, int)
        ) and self.selector < len(node):
            return [str(self.selector)], [node[self.selector]]
        
        return None, None
    
    def _match_key(
        self,
        node: Data,
    ):
        if not isinstance(node, CommentedMap):
            return None, None
        
        if value := node.get(self.selector):
            return [self.selector], [value]
        
        return None, None
    
    def _match_pattern(
        self,
        node: Data,
    ):
        
        if not isinstance(self.selector, re.Pattern):
            return None, None
        
        if isinstance(node, dict):
            keys = [
                key for key in node.keys()
                if self.selector.match(key)
            ]

            return keys, [
                node.get(key) for key in keys
            ]
            
        return None, None
    
    def _match_pattern_range(
        self,
        node: Data,
    ):
        if not isinstance(node, list) or not isinstance(self.selector, re.Pattern):
            return None, None
        
        matches = [
            (idx, item)
            for idx, item in enumerate(node)
            if (
                isinstance(item, (dict, list))
                and any([
                    self.selector.match(str(val))
                    for val in item
                ])
            ) or (
                isinstance(item, str)
                and self.selector.match(str(item))
            )
        ]
        
        return (
            [idx for idx, _ in matches],
            [item for _, item in matches]
        )
    
    def _match_unbound_range(
        self,
        node: Data,
    ):
        if not isinstance(node, list):
            return None, None

        return (
            ['[]'],
            [node],
        )
    
    def _match_key_range(
        self,
        node: Data,
    ):
        if not isinstance(node, CommentedSeq):
            return None, None
        
        matches = [
            (
                idx,
                value
            ) for idx, value in enumerate(node) if (
                str(value) == self.selector
            ) or (
                isinstance(value, (dict, list))
                and self.selector in value
            )
        ]

        for idx, match in enumerate(matches):
            match_idx, item = match

            if isinstance(item, dict):
                matches[idx] = (
                    match_idx,
                    item.get(self.selector),
                )

            elif isinstance(item, list):
                match_idx[idx] = (
                    match_idx,
                    matches[match_idx],
                )

        return (
            [str(idx) for idx, _ in matches],
            [item for _, item in matches]
        )
    
    def _match_nested_map(
        self,
        node: Data,
    ):
        if not isinstance(node, dict):
            return None, None
        
        keys: list[str] = []
        found: list[Data] = []

        for value in node.values():
            if isinstance(value, list):
                nested_keys, nested_found = self._match_nested(value)
                keys.extend([
                    f'[[{key}]]'
                    for key in nested_keys
                ])
                found.extend(nested_found)

            elif isinstance(value, dict):
                nested_keys, nested_found = self._match_nested(value)
                keys.extend([
                    f'(({key}))'
                    for key in nested_keys
                ])
                found.extend(nested_found)

            return (
                keys,
                found,
            )

    def _match_nested_range(
        self,
        node: Data
    ):
        if not isinstance(node, list):
            return None, None
        
        keys: list[str] = []
        found: list[Data] = []

        for item in node:
            if isinstance(item, list):
                nested_keys, nested_found = self._match_nested(item)
                keys.extend([
                    f'[[{key}]]'
                    for key in nested_keys
                ])
                found.extend(nested_found)

            elif isinstance(item, dict):
                nested_keys, nested_found = self._match_nested(item)
                keys.extend([
                    f'(({key}))'
                    for key in nested_keys
                ])
                found.extend(nested_found)

        return (
            keys,
            found,
        )
    
    def _match_nested(
        self,
        node: Data,
    ): 
        found: Items = deque()
        keys: Deque[str] = deque()

        for token in self._nested:
            matched_keys, matches = token.match(node)

            if matched_keys and matches:
                keys.extend(matched_keys)
                found.extend(matches)

        return keys, found
    
    def _match_value(
        self,
        node: Data
    ):
        
        if not isinstance(node, CommentedMap):
            return None, None

        keys: list[str] = []
        found: list[Any] = []
        for group in self._selector_segments:
            for segment in group:

                operator = ValueOperator(segment)

                match_keys: list[str] = []
                match_found: list[Any] = []


                if match := operator.match(node):
                    key, value = match
                    match_keys.append(key)
                    match_found.append(value)

                if len(match_keys) == len(group):
                    keys.append(
                        '&'.join(match_keys)
                    )
                    found.append(node)
                    break

        if len(keys) < 1 or len(found) < 1:
            return None, None

        return keys, found


import asyncio
import sys
from cfn_check.yaml import YAML
from cfn_check.yaml.comments import CommentedBase

async def write_to_stdout(data: CommentedBase):
    loop = asyncio.get_event_loop()

    yaml = YAML(typ=['rt'])
    yaml.preserve_quotes = True
    yaml.width = 4096
    yaml.indent(mapping=2, sequence=4, offset=2)
    await loop.run_in_executor(
        None,
        yaml.dump,
        data,
        sys.stdout,
    )

async def write_multiple_files_to_stdout(data: CommentedBase):
    loop = asyncio.get_event_loop()

    yaml = YAML(typ=['rt'])
    yaml.preserve_quotes = True
    yaml.width = 4096
    yaml.indent(mapping=2, sequence=4, offset=2)
    await loop.run_in_executor(
        None,
        yaml.dump_all,
        data,
        sys.stdout,
    )


from cfn_check.collection.collection import Collection
from cfn_check.validation.validator import Validator


def bind(
    rule_set: Collection,
    validation: Validator,
):
    validation.func = validation.func.__get__(rule_set, rule_set.__class__)
    setattr(rule_set, validation.func.__name__, validation.func)

    return validation



from typing import Any
from cfn_check.yaml.comments import CommentedMap, CommentedSeq


def assign(parent: CommentedMap | CommentedSeq | None, key_or_index: Any, value: Any):
    if parent is None:
        return  # root already set
    if isinstance(parent, CommentedMap):
        parent[key_or_index] = value
    else:
        # key_or_index is an int for sequences
        # Ensure sequence large enough (iterative approach assigns in order, so append is fine)
        parent.append(value)


import re
import sys
from .token import Token
from .token_type import TokenType


class QueryParser:

    def __init__(self):
        self.numbers_pattern = re.compile(r'\d+')

    def parse(
        self,
        query: str,
    ):
        
        tokens: list[Token] = []

        if query.startswith('[') and query.endswith(']'):
            tokens.extend(
                self._parse_range_selector_token(query),
            )

        elif query.startswith('<') and query.endswith('>'):
            tokens.append(
                Token(
                    re.compile(query),
                    TokenType.PATTERN,
                )
            )

        elif query.startswith('(') and query.endswith(')'):
            tokens.append(
                Token(
                    query[1:-1],
                    TokenType.VALUE_MATCH,
                )
            )
        elif query == "*":
            tokens.append(
                Token(
                    query,
                    TokenType.WILDCARD,
                )
            )

        else:
            tokens.append(
                Token(
                    query,
                    TokenType.KEY,
                )
            )

        return tokens

    def _parse_range_selector_token(
        self,
        query: str,
    ):
        segments = [
            segment
            for segment in query[1:-1].split(',')
            if len(segment) > 0
        ]
        tokens: list[Token] = []

        if query.startswith('(') and query.endswith(')'):
            tokens.append(
                Token(
                    query[1:-1],
                    TokenType.VALUE_MATCH,
                )
            )

        if len(segments) < 1:
            tokens.append(
                Token(
                    None,
                    TokenType.UNBOUND_RANGE,
                ),
            )

        else:  
            tokens.extend([
                token
                for segment in segments
                for token in self._parse_selector_segment(segment)
            ])

        return tokens
    
    def _parse_selector_segment(self, segment: str):
        
        if segment.startswith('[') and segment.endswith(']'):
            tokens = self._parse_range_selector_token(segment)

            return [
                Token(
                    segment,
                    TokenType.NESTED_RANGE,
                    nested=tokens,
                ),
            ]
        
        elif segment.startswith('<') and segment.endswith('>'):
            return [
                Token(
                    re.compile(segment),
                    TokenType.PATTERN_RANGE,

                )
            ]
        
        elif segment == '*':
            return [
                Token(
                    segment,
                    TokenType.WILDCARD_RANGE,
                )
            ]
        
        elif '-' in segment:
            return [
                 self._parse_bound_range(
                    segment.split('-', maxsplit=1)
                )
            ]
        
        elif '=' in segment:
            return [
                Token(
                    segment,
                    TokenType.VALUE_MATCH
                )
            ]
        
        elif match := self.numbers_pattern.match(segment):
                return [
                    Token(
                        int(match.group(0)),
                        TokenType.INDEX,
                    )
                ]
        
        else:
            return [
                Token(
                    segment,
                    TokenType.KEY_RANGE
                )
            ]

    def _parse_bound_range(self, segment: tuple[str, ...]):
   
        start, stop = segment
        if not self.numbers_pattern.match(start):
            start = 0

        if not self.numbers_pattern.match(stop):
            stop = str(sys.maxsize)

        return Token(
            (
                int(start),
                int(stop),
            ),
            TokenType.BOUND_RANGE,
        )



from __future__ import annotations
import base64
import json
import re
from typing import Callable, Any
from collections import deque
from cfn_check.yaml.tag import Tag
from cfn_check.yaml.comments import TaggedScalar, CommentedMap, CommentedSeq
from .cidr_solver import IPv4CIDRSolver
from .utils import assign

from cfn_check.shared.types import (
    Data,
    Items,
    YamlObject,
)

Resolver = Callable[
    [
        CommentedMap,
        CommentedMap | CommentedSeq | TaggedScalar | YamlObject
    ],
    CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
]

class Renderer:

    def __init__(self):
        self.items: Items = deque()
        self._sub_pattern = re.compile(r'\$\{([\w+::]+)\}')
        self._sub_inner_text_pattern = re.compile(r'[\$|\{|\}]+')
        self._visited: list[str | int] = []
        self._data: YamlObject = {}
        self._parameters = CommentedMap()
        self._mappings = CommentedMap()
        self._parameters_with_defaults: dict[str, str | int | float | bool | None] = {}
        self._selected_mappings = CommentedMap()
        self._conditions = CommentedMap()
        self._references: dict[str, str] = {}
        self._resources: dict[str, YamlObject] = CommentedMap()
        self._attributes: dict[str, str] = {}
        self._import_values: dict[
            str,
            CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
        ] = {}
        self._availability_zones = CommentedSeq()

        self._inline_functions = {
            'Fn::ForEach': re.compile(r'Fn::ForEach::\w+'),
            'Fn::If': re.compile(r'Fn::If'),
            'Fn::And': re.compile(r'Fn::And'),
            'Fn::Equals': re.compile(r'Fn::Equals'),
            'Fn::Not': re.compile(r'Fn::Not'),
            'Fn::Or': re.compile(r'Fn::Or'),
            'Fn:GetAtt': re.compile(r'Fn::GetAtt'),
            'Fn::Join': re.compile(r'Fn::Join'),
            'Fn::Sub': re.compile(r'Fn::Sub'),
            'Fn::Base64': re.compile(r'Fn::Base64'),
            'Fn::Split': re.compile(r'Fn::Split'),
            'Fn::Select': re.compile(r'Fn::Select'),
            'Fn::ToJsonString': re.compile(r'Fn::ToJsonString'),
            'Fn::Condition': re.compile(r'Fn::Condition'),
            'Fn::Cidr': re.compile(r'Fn::Cidr'),
            'Fn::Length': re.compile(r'Fn::Length'),
            'Fn::GetAZs': re.compile(r'Fn::GetAZs'),
            'Fn::ImportValue': re.compile(r'Fn::ImportValue'),
        }

        self._inline_resolvers = {
            'Fn::ForEach': self._resolve_foreach,
            'Fn::If': self._resolve_if,
            'Fn::And': self._resolve_and,
            'Fn::Equals': self._resolve_equals,
            'Fn::Not': self._resolve_not,
            'Fn::Or': self._resolve_or,
            'Fn:GetAtt': self._resolve_getatt,
            'Fn::Join': self._resolve_join,
            'Fn::Sub': self._resolve_sub,
            'Fn::Base64': self._resolve_base64,
            'Fn::Split': self._resolve_split,
            'Fn::Select': self._resolve_select,
            'Fn::ToJsonString': self._resolve_tree_to_json,
            'Fn::Condition': self._resolve_condition,
            'Fn::Cidr': self._resolve_cidr,
            'Fn::Length': self._resolve_length,
            'Fn::GetAZs': self._resolve_get_availability_zones,
            'Fn::ImportValue': self._resolve_import_value,
        }

        self._resolvers: dict[str, Callable[[CommentedMap, str], YamlObject]] = {
            '!Ref': self._resolve_ref,
            '!FindInMap': self._resolve_by_subset_query,
            '!GetAtt': self._resolve_getatt,
            '!Join': self._resolve_join,
            '!Sub': self._resolve_sub,
            '!Base64': self._resolve_base64,
            '!Split': self._resolve_split,
            '!Select': self._resolve_select,
            '!ToJsonString': self._resolve_tree_to_json,
            '!Equals': self._resolve_equals,
            '!If': self._resolve_if,
            '!Condition': self._resolve_condition,
            '!And': self._resolve_and,
            '!Not': self._resolve_not,
            '!Or': self._resolve_or,
            '!Cidr': self._resolve_cidr,
            '!GetAZs': self._resolve_get_availability_zones,
            '!ImportValue': self._resolve_import_value,
        }

    def render(
        self,
        template: YamlObject,
        attributes: dict[str, Any] | None = None,
        availability_zones: list[str] | None = None,
        import_values: dict[str, tuple[str, CommentedMap]] | None = None,
        mappings: dict[str, str] | None = None,
        parameters: dict[str, Any] | None = None,
        references: dict[str, str] | None = None,
    ):

        self._sources = list(template.keys())

        self._assemble_parameters(template)

        if attributes:
            self._attributes = self._process_attributes(attributes)

        if availability_zones:
            self._availability_zones = CommentedSeq(availability_zones)

        if import_values:
            for _, (import_key, imported_template) in import_values.items():
                self._import_values[import_key] = self._resolve_external_export(import_key, imported_template)

        self._mappings = template.get('Mappings', CommentedMap())
        if mappings:
            self._selected_mappings = self._assemble_mappings(mappings)

        self._parameters = template.get('Parameters', CommentedMap())
        if parameters:
            self._parameters_with_defaults.update(parameters)

        if references:
            self._references.update(references)

        self._resources = template.get('Resources', CommentedMap())
        self._conditions = template.get('Conditions', CommentedMap())

        return self._resolve_tree(template)

    def _resolve_tree(self, root: YamlObject):
        self.items.clear()
        self.items.append((None, None, root))

        while self.items:
            parent, accessor, node = self.items.pop()
            if match := self._match_and_resolve_accessor_fn(
                root,
                parent,
                accessor,
                node,
            ):
                root.update(match)
            
            if isinstance(node, TaggedScalar):
                # Replace in parent
                if parent is not None and (
                    resolved := self._resolve_tagged(root, node)
                ):
                    parent[accessor] = resolved

            elif isinstance(node, CommentedMap):
                if isinstance(node.tag, Tag) and node.tag.value is not None and parent:
                    resolved_node = self._resolve_tagged(root, node)
                    parent[accessor] = resolved_node

                elif isinstance(node.tag, Tag) and node.tag.value is not None:
                    node = self._resolve_tagged(root, node)
                    for k in reversed(list(node.keys())):
                        self.items.append((node, k, node[k]))

                    root = node
                
                else:
                    # Process keys in reverse order for proper DFS
                    for k in reversed(list(node.keys())):
                        self.items.append((node, k, node[k]))

            elif isinstance(node, CommentedSeq):
                
                if isinstance(node.tag, Tag) and node.tag.value is not None and parent:
                    resolved_node = self._resolve_tagged(root, node)
                    parent[accessor] = resolved_node

                elif isinstance(node.tag, Tag) and node.tag.value is not None:
                    node = self._resolve_tagged(root, node)
                    
                    for idx, val in enumerate(reversed(node)):
                        self.items.append((node, idx, val))

                    root = node

                else:
                    # Process indices in reverse order for proper DFS
                    for idx, val in enumerate(reversed(node)):
                        self.items.append((node, idx, val))

        return root
    
    def _match_and_resolve_accessor_fn(
        self,
        root: CommentedMap,
        parent: CommentedMap | CommentedSeq | TaggedScalar | YamlObject | None,
        accessor: str | int | None,
        node: CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
    ):
        if not isinstance(accessor, str):
            return None
        
        resolver: Resolver | None  = None
        matcher_pattern: re.Pattern | None = None
        for key, pattern in self._inline_functions.items():
            if pattern.match(accessor):
                matcher_pattern = pattern
                resolver = self._inline_resolvers[key]

        if resolver is None:
            return None
        
        result = resolver(root, node)

        return self._replace_target(
            root,
            parent,
            result,
            matcher_pattern,
        )

    def _resolve_tagged(self, root: CommentedMap, node: TaggedScalar | CommentedMap | CommentedSeq):
        resolver: Callable[[CommentedMap, str], YamlObject] | None = None
        
        if isinstance(node.tag, Tag) and (
            resolver := self._resolvers.get(node.tag.value)
        ):    
            return resolver(root, node)
    
    def _resolve_ref(self, root: YamlObject, scalar: TaggedScalar):
        '''
        Sometimes we can resolve a !Ref if it has an explicit correlation
        to a Resources key or input Parameter. This helps reduce the amount
        of work we have to do when resolving later.
        '''
        if val := self._parameters_with_defaults.get(scalar.value):
            return val
        
        elif scalar.value in self._parameters:
            return scalar

        elif scalar.value in self._resources:
            return scalar.value
        
        elif ref := self._references.get(scalar.value):
            return ref

        else:
            return self._resolve_subtree(
                root,
                self._find_matching_key(root, scalar.value),
            )
    
    def _resolve_getatt(
        self,
        root: CommentedMap, 
        query: TaggedScalar | CommentedMap | CommentedSeq,
    ) -> YamlObject | None:
        steps: list[str] = []
        
        if isinstance(query, TaggedScalar):
            steps_string: str = query.value
            steps = steps_string.split('.')

        elif (
            resolved := self._resolve_subtree(root, query)
        ) and isinstance(
            resolved,
            CommentedSeq,
        ):
            steps = resolved

        if value := self._attributes.get(
            '.'.join(steps)
        ):
            return value

        current = self._resources.get(steps[0], CommentedMap()).get(
            'Properties',
            CommentedMap(),
        )

        for step in steps[1:]:
            if step == 'Value':
                return current
            # Mapping
            if isinstance(current, (CommentedMap, dict)):
                if step in current:
                    current = current[step]
                else:
                    return None
            # Sequence
            elif isinstance(current, (CommentedSeq, list)):
                try:
                    idx = int(step)
                except ValueError:
                    return None
                if 0 <= idx < len(current):
                    current = current[idx]
                else:
                    return None
            else:
                # Hit a scalar (including TaggedScalar) before consuming all steps
                return None
        
        return current
    
    def _resolve_join(
        self,
        root: CommentedMap,
        source: CommentedSeq,
    ) -> Any:
        if len(source) < 2:
            return ''
        
        delimiter = source[0]
        if isinstance(delimiter, (TaggedScalar, CommentedMap, CommentedSeq)):
            delimiter = str(self._resolve_tagged(root, delimiter))

        else:
            delimiter = str(delimiter)  
        
        subselction = source[1:]
        resolved = self._resolve_subtree(root, subselction)

        if not isinstance(resolved, CommentedSeq):
            return source

        return delimiter.join([
            str(self._resolve_tagged(
                root,
                node,
            ))
            if isinstance(
                node,
                (TaggedScalar, CommentedMap, CommentedSeq)
            ) else node 
            for subset in resolved
            for node in subset
        ])
    
    def _resolve_sub(
        self, 
        root: CommentedMap,
        source: CommentedSeq | TaggedScalar,
    ):
        if isinstance(source, TaggedScalar) and isinstance(
            source.tag,
            Tag,
        ):
            source_string = source.value
            variables = self._resolve_template_string(source_string)
            return self._resolve_sub_ref_queries(
                variables,
                source_string,
            )

        elif len(source) > 1:
            source_string: str = source[0]
            template_vars = self._resolve_template_string(source_string)
            variables = source[1:]
            resolved: list[dict[str, Any]] = self._resolve_subtree(root, variables)
            
            for resolve_var in resolved:
                for template_var, accessor in template_vars:
                    if val := resolve_var.get(accessor):
                        source_string = source_string.replace(template_var, val)

            return source_string

        return source
    
    def _resolve_base64(
        self,
        root: CommentedMap,
        source: CommentedMap | CommentedSeq | TaggedScalar,
    ):
        if isinstance(source, TaggedScalar) and isinstance(
            source.tag, 
            Tag,
        ) and isinstance(
            source.tag.value,
            str,
        ):
            return base64.b64encode(source.tag.value.encode()).decode('ascii')
        
        elif (
            resolved := self._resolve_subtree(root, source)
        ) and isinstance(
            resolved,
            str
        ):
          return  base64.b64encode(resolved.encode()).decode('ascii')
        
        return source
    
    def _resolve_foreach(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if not isinstance(source, CommentedSeq) or len(source) < 3:
            return source
        
        identifier = source[0]
        if not isinstance(identifier, str):
            identifier = self._resolve_subtree(root, identifier)

        collection = source[1]
        if not isinstance(collection, list):
            return source
        
        collection: list[str] = self._resolve_subtree(root, collection)

        output = source[2]
        if not isinstance(output, CommentedMap):
            return source

        resolved_items = CommentedMap()
        for item in collection:
            self._references[identifier] = item
            resolved_items.update(
                self._resolve_foreach_item(
                    root,
                    self._copy_subtree(output),
                ) 
            )
        
        return resolved_items
    
    def _resolve_foreach_item(
        self,
        root: CommentedMap,
        output_item: CommentedMap,
    ):
        output_map: dict[str, CommentedMap] = {}
        for output_key, output_value in output_item.items():
            variables = self._resolve_template_string(output_key)
            resolved_key = self._resolve_sub_ref_queries(
                variables,
                output_key,
            )

            output_map[resolved_key] = self._resolve_subtree(
                root,
                output_value,
            )

        return output_map
    
    def _resolve_split(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ) or len(source) != 2:
            return source
        
        delimiter = source[0]
        if not isinstance(
            delimiter,
            str,
        ):
            delimiter = self._resolve_subtree(root, delimiter)

        target = source[1]
        if not isinstance(
            target,
            str,
        ):
            target = self._resolve_subtree(root, target)

        if isinstance(delimiter, str) and isinstance(target, str):
            return CommentedSeq(target.split(delimiter))
        
        return target
    
    def _resolve_select(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ) or len(source) != 2:
            return source
        
        
        index = source[0]
        if not isinstance(
            index,
            int,
        ):
            index = self._resolve_subtree(root, index)

        target = self._resolve_subtree(root, source[1])
        if index > len(target):
            return source
        
        return target[index]
    
    def _resolve_equals(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ) or len(source) != 2:
            return source
        
        item_a = source[0]
        if isinstance(
            item_a,
            (CommentedMap, CommentedSeq, TaggedScalar),
        ):
            item_a = self._resolve_subtree(root, item_a)

        item_b = source[1]
        if isinstance(
            item_b,
            (CommentedMap, CommentedSeq, TaggedScalar),
        ):
            item_b = self._resolve_subtree(root, item_b)

        return item_a == item_b

    def _resolve_if(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ) or len(source) != 3:
            return source
        
        condition_key = source[0]
        if isinstance(
            condition_key,
            (CommentedMap, CommentedSeq, TaggedScalar),
        ):
            condition_key = self._resolve_subtree(root, condition_key)

        result = self._resolve_subtree(root, self._conditions.get(condition_key))

        true_result = source[1]
        if isinstance(
            true_result,
            (CommentedMap, CommentedSeq, TaggedScalar),
        ):
            true_result = self._resolve_subtree(root, true_result)

        false_result = source[2]
        
        return true_result if isinstance(result, bool) and result else false_result
    
    def _resolve_condition(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, CommentedSeq),
        ):
            return source
        
        if (
            condition := self._conditions.get(source.value)
        ) and isinstance(
            condition,
            (CommentedMap, CommentedSeq, TaggedScalar)
        ) and (
            result := self._resolve_subtree(root, condition)
        ) and isinstance(
            result,
            bool,
        ):
            return result
        
        elif (
            condition := self._conditions.get(source.value)
        ) and isinstance(
            condition,
            bool,
        ):
            return condition
        
        return source
    
    def _resolve_and(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ):
            return source
        
        resolved = self._resolve_subtree(root, CommentedSeq([
            item for item in source
        ]))
        if not isinstance(resolved, CommentedSeq):
            return source
        
    
        for node in resolved:
            if not isinstance(node, bool):
                return source
        
        return all(resolved)
    
    def _resolve_not(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ):
            return source
        
        resolved = self._resolve_subtree(root, CommentedSeq([
            item for item in  source
        ]))
        if not isinstance(resolved, CommentedSeq):
            return source
        
        for node in resolved:
            if not isinstance(node, bool):
                return source
        
        return not all(resolved)
    
    def _resolve_or(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if isinstance(
            source,
            (CommentedMap, TaggedScalar),
        ):
            return source
        
        resolved = self._resolve_subtree(root, CommentedSeq([
            item for item in source
        ]))
        if not isinstance(resolved, CommentedSeq):
            return source
        
    
        for node in resolved:
            if not isinstance(node, bool):
                return source
        
        return any(resolved)
    
    def _resolve_cidr(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        if not isinstance(
            source,
            CommentedSeq,
        ) or len(source) < 3:
            return source
        
        cidr = self._resolve_subtree(root, source[0])
        if not isinstance(cidr, str):
            return source

        subnets_requested = source[1]
        subnet_cidr_bits = source[2]

        ipv4_solver = IPv4CIDRSolver(
            cidr,
            subnets_requested,
            subnet_cidr_bits,
        )

        return CommentedSeq(ipv4_solver.provision_subnets())

    def _resolve_tree_to_json(
        self,
        root: CommentedMap,
        source: CommentedSeq | CommentedMap | TaggedScalar,
    ):
        
        stack: list[tuple[CommentedMap | CommentedSeq | None, Any | None, Any]] = [(None, None, source)]

        while stack:
            parent, accessor, node = stack.pop()
            if isinstance(node, TaggedScalar):
                # Replace in parent
                if parent is not None and (
                    resolved := self._resolve_tagged(root, node)
                ):
                    parent[accessor] = resolved

                elif (
                    resolved := self._resolve_tagged(root, node)
                ):
                    source = resolved

            elif isinstance(node, CommentedMap):
                if isinstance(node.tag, Tag) and node.tag.value is not None and parent and (
                    resolved_node := self._resolve_tagged(root, node)
                ) and node != source:
                    parent[accessor] = resolved_node

                elif isinstance(node.tag, Tag) and node.tag.value is not None and node != source:
                    node = self._resolve_tagged(root, node)
                    for k in reversed(list(node.keys())):
                        stack.append((node, k, node[k]))

                    source = node
                
                else:
                    # Push children (keys) in reverse for DFS order
                    for k in reversed(list(node.keys())):
                        stack.append((node, k, node[k]))

            elif isinstance(node, CommentedSeq):
                if isinstance(node.tag, Tag) and node.tag.value is not None and parent and (
                    resolved_node := self._resolve_tagged(root, node)
                )  and node != source :
                    parent[accessor] = resolved_node

                elif isinstance(node.tag, Tag) and node.tag.value is not None and node != source:
                    node = self._resolve_tagged(root, node)
                    for idx, val in enumerate(reversed(node)):
                        stack.append((node, idx, val))

                    source = node
                
                else:
                    # Process indices in reverse order for proper DFS
                    for idx, val in enumerate(reversed(node)):
                        stack.append((node, idx, val))
        
        return json.dumps(source)
    
    def _resolve_length(
        self,
        root: CommentedMap,
        source: CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
    ):
        items = CommentedSeq()
        if isinstance(source, TaggedScalar):
            items = self._resolve_tagged(root, source)

        elif isinstance(source, (CommentedMap, CommentedSeq)):
            items = self._resolve_subtree(root, source)

        elif isinstance(source, (str, list, dict)):
            items = source

        else:
            return source
        
        return len(items)
    
    def _resolve_get_availability_zones(
        self,
        _: CommentedMap,
        source: CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
    ):
        if not isinstance(source, TaggedScalar):
            return source
        
        return self._availability_zones
    
    def _resolve_import_value(
        self,
        _: CommentedMap,
        source: CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
    ): 
        if not isinstance(source, TaggedScalar):
            return source
        
        return self._import_values.get(source.value)

    def _resolve_external_export(
        self, 
        key: str,
        template: CommentedMap,
    ):
        outputs: CommentedMap = template.get('Outputs', CommentedMap())
        exports: CommentedMap = outputs.get('Export', CommentedMap())

        if subtree := exports.get(key):
            return self._resolve_subtree(template, subtree)
        
        return None

    def _copy_subtree(
        self,
        root: CommentedMap | CommentedSeq | TaggedScalar | YamlObject,
    ) -> Any:
        """
        Depth-first clone of a ruamel.yaml tree.
        - Rebuilds CommentedMap/CommentedSeq
        - Copies TaggedScalar (preserves tag and value)
        - Scalars are copied as-is
        Note: does not preserve comments/anchors.
        """
        if isinstance(root, CommentedMap):
            root_clone: Any = CommentedMap()
        elif isinstance(root, CommentedSeq):
            root_clone = CommentedSeq()
        elif isinstance(root, TaggedScalar):
            return TaggedScalar(
                value=root.value,
                tag=root.tag,
            )
        else:
            return root

        stack: list[
            tuple[
                Any,
                CommentedMap | CommentedSeq | None,
                Any | None,
            ]
        ] = [(root, None, None)]

        built: dict[
            int,
            CommentedMap | CommentedSeq,
        ] = {id(root): root_clone}

        while stack:
            in_node, out_parent, out_key = stack.pop()

            if isinstance(in_node, CommentedMap):
                out_container = built.get(id(in_node))
                if out_container is None:
                    out_container = CommentedMap()
                    built[id(in_node)] = out_container
                assign(out_parent, out_key, out_container)

                for k in reversed(list(in_node.keys())):
                    v = in_node[k]
                    if isinstance(v, CommentedMap):
                        child = CommentedMap()
                        built[id(v)] = child

                        stack.append((v, out_container, k))
                    elif isinstance(v, CommentedSeq):
                        child = CommentedSeq()
                        built[id(v)] = child

                        stack.append((v, out_container, k))
                    elif isinstance(v, TaggedScalar):
                        ts = TaggedScalar(
                            value=v.value,
                            tag=v.tag,
                        )

                        out_container[k] = ts
                    else:
                        out_container[k] = v

            elif isinstance(in_node, CommentedSeq):
                out_container = built.get(id(in_node))
                if out_container is None:
                    out_container = CommentedSeq()
                    built[id(in_node)] = out_container
                assign(out_parent, out_key, out_container)

                for idx in reversed(range(len(in_node))):
                    v = in_node[idx]

                    if isinstance(v, CommentedMap):
                        child = CommentedMap()
                        built[id(v)] = child

                        stack.append((v, out_container, idx))
                    elif isinstance(v, CommentedSeq):
                        child = CommentedSeq()
                        built[id(v)] = child

                        stack.append((v, out_container, idx))
                    elif isinstance(v, TaggedScalar):
                        ts = TaggedScalar(
                            value=v.value,
                            tag=v.tag,
                        )

                        out_container.append(ts)
                    else:
                        out_container.append(v)

            elif isinstance(in_node, TaggedScalar):
                ts = TaggedScalar(
                    value=in_node.value,
                    tag=in_node.tag,
                )

                assign(out_parent, out_key, ts)

            else:
                assign(out_parent, out_key, in_node)

        return root_clone

    def _replace_target(
        self,
        root: CommentedMap,
        target: CommentedMap,
        replacement: Any,
        matcher_pattern: re.Pattern
    ) -> CommentedMap: 
        if not isinstance(target, CommentedMap):
            return root

        if root is target:
            return replacement
        
        stack: list[tuple[Any, Any | None, Any | None]] = [(root, None, None)]
        
        while stack:
            node, parent, accessor = stack.pop()
            
            if isinstance(node, CommentedMap):
                for k in reversed(list(node.keys())):
                    child = node[k]
                    if child is target and isinstance(child, CommentedMap):
                        for key in list(target.keys()):
                            if matcher_pattern.match(key):
                                del child[key]

                        if isinstance(replacement, CommentedMap):
                            child.update(replacement)
                            node[k] = child
                            
                        else:
                            node[k] = replacement

                        if parent:
                            parent[accessor] = node

                        return root
                    
                    stack.append((child, node, k))
            
            elif isinstance(node, CommentedSeq):
                for idx in reversed(range(len(node))):
                    child = node[idx]
                    if child is target and isinstance(child, CommentedMap):
                        for key in list(target.keys()):
                            if matcher_pattern.match(key):
                                del child[key]

                        if isinstance(replacement, CommentedMap):
                            child.update(replacement)
                            node[idx] = child
                            
                        else:
                            node[idx] = replacement

                        if parent:
                            parent[accessor] = node

                        return root
                    
                    stack.append((child, node, idx))
        
        return root

    def _resolve_subtree(
        self,
        root: CommentedMap,
        source: CommentedSeq
    ) -> Any:
        """
        Iterative DFS over a ruamel.yaml tree.
        - CommentedMap/CommentedSeq are traversed.
        """
        stack: list[tuple[CommentedMap | CommentedSeq | None, Any | None, Any]] = [(None, None, source)]
        
        source_parent, source_index = self._find_parent(root, source)
        
        while stack:
            parent, accessor, node = stack.pop()
            if match := self._match_and_resolve_accessor_fn(
                    root, 
                    parent, 
                    accessor,
                    node,
            ):
                root.update(match)
                # At this point we've likely (and completely)
                # successfully nuked the source from orbit
                # so we need to fetch it from the source parent
                # to get it back (i.e. the ref is no longer 
                # correct).
                source = source_parent[source_index]

            if isinstance(node, TaggedScalar):
                # Replace in parent
                if parent is not None and (
                    resolved := self._resolve_tagged(root, node)
                ):
                    parent[accessor] = resolved

                elif (
                    resolved := self._resolve_tagged(root, node)
                ):
                    source = resolved

            elif isinstance(node, CommentedMap):
                if isinstance(node.tag, Tag) and node.tag.value is not None and parent:
                    resolved_node = self._resolve_tagged(root, node)
                    parent[accessor] = resolved_node

                elif isinstance(node.tag, Tag) and node.tag.value is not None:
                    node = self._resolve_tagged(root, node)
                    for k in reversed(list(node.keys())):
                        stack.append((node, k, node[k]))

                    source = node
                
                else:
                    # Push children (keys) in reverse for DFS order
                    for k in reversed(list(node.keys())):
                        stack.append((node, k, node[k]))

            elif isinstance(node, CommentedSeq):
                if isinstance(node.tag, Tag) and node.tag.value is not None and parent:
                    resolved_node = self._resolve_tagged(root, node)
                    parent[accessor] = resolved_node

                elif isinstance(node.tag, Tag) and node.tag.value is not None:
                    node = self._resolve_tagged(root, node)
                    for idx, val in enumerate(reversed(node)):
                        stack.append((node, idx, val))

                    source = node
                
                else:
                    # Process indices in reverse order for proper DFS
                    for idx, val in enumerate(reversed(node)):
                        stack.append((node, idx, val))
        
        return source
    
    def _resolve_by_subset_query(
        self, 
        root: CommentedMap, 
        subset: CommentedMap | CommentedSeq,
    ) -> YamlObject | None:
        """
        Traverse `subset` iteratively. For every leaf (scalar or TaggedScalar) encountered in `subset`,
        use its value as the next key/index into `root`. Return (path, value) where:
        - path: list of keys/indices used to reach into `root`
        - value: the value at the end of traversal, or None if a step was missing (early return)
        TaggedScalar is treated as a leaf and its .value is used as the key component.
        """
        current = self._mappings
        path = []

        stack = [(subset, [])]
        while stack:
            node, _ = stack.pop()

            if isinstance(node, CommentedMap):

                if isinstance(node.tag, Tag) and node.tag.value is not None and (
                    node != subset
                ):
                    resolved_node = self._resolve_tagged(root, node)
                    stack.append((resolved_node, []))
                
                else:
                    for k in reversed(list(node.keys())):
                        stack.append((node[k], []))

            elif isinstance(node, CommentedSeq):

                if isinstance(node.tag, Tag) and node.tag.value is not None and (
                    node != subset
                ):
                    resolved_node = self._resolve_tagged(root, node)
                    stack.append((resolved_node, []))

                else:
                    for val in reversed(node):
                        stack.append((val, []))
            else:
                # Leaf: scalar or TaggedScalar
                key = self._resolve_tagged(
                    self._selected_mappings,
                    node,
                ) if isinstance(node, TaggedScalar) else node
                path.append(key)

                if isinstance(current, CommentedMap):
                    if key in current:
                        current = current[key]
                    else:
                        return None
                elif isinstance(current, CommentedSeq) and isinstance(key, int) and 0 <= key < len(current):
                    current = current[key]
                else:
                    return None
                
        if isinstance(current, TaggedScalar):
            return path, self._resolve_tagged(
                self._selected_mappings,
                current,
            )

        return current
    
    def _find_matching_key(
        self,
        root: CommentedMap, 
        search_key: str,
    ):
        """Returns the first path (list of keys/indices) to a mapping with key == search_key, and the value at that path."""
        stack = [(root, [])]
        while stack:
            node, path = stack.pop()
            if isinstance(node, CommentedMap):
                for k in reversed(list(node.keys())):
                    if k == search_key:
                        return node[k]
                    stack.append((node[k], path + [k]))
            elif isinstance(node, CommentedSeq):
                for idx, item in reversed(list(enumerate(node))):
                    stack.append((item, path + [idx]))

        return None  # No match found
    
    def _find_parent(
        self,
        root: CommentedMap,
        target: CommentedMap,
    ) -> CommentedMap: 
        
        stack: list[tuple[Any, Any | None, Any | None]] = [(root, None, None)]
        
        while stack:
            node, parent, accessor = stack.pop()
            
            if isinstance(node, CommentedMap):
                for k in reversed(list(node.keys())):
                    child = node[k]
                    if child is target and isinstance(child, CommentedMap):
                        return node, k
                    
                    stack.append((child, node, k))
            
            elif isinstance(node, CommentedSeq):
                for idx in reversed(range(len(node))):
                    child = node[idx]
                    if child is target and isinstance(child, CommentedMap):
                        return node, node.index(child)
                    
                    stack.append((child, node, idx))
        
        return None, None
    
    def _assemble_parameters(self, resources: YamlObject):
        params: dict[str, Data] = resources.get("Parameters", {})
        for param_name, param in params.items():
            if isinstance(param, CommentedMap) and (
                default := param.get("Default")
            ):
                self._parameters_with_defaults[param_name] = default

    def _assemble_mappings(self, mappings: dict[str, str]):
        for mapping, value in mappings.items():
            if (
                map_data := self._mappings.get(mapping)
            ) and (
                selected := map_data.get(value)
            ):
                self._selected_mappings[mapping] = selected

    def _process_attributes(
        self,
        attributes: dict[str, Any],
    ):
        return {
            key: self._process_python_structure(value)
            for key, value in attributes.items()
        }

    def _process_python_structure(
        self,
        obj: Any
    ) -> Any:
        """
        Convert arbitrarily nested Python data (dict/list/scalars) into ruamel.yaml
        CommentedMap/CommentedSeq equivalents using iterative DFS. Scalars are returned as-is.
        """
        # Fast path for scalars
        if not isinstance(obj, (dict, list)):
            return obj

        # Create root container
        if isinstance(obj, dict):
            root_out: Any = CommentedMap()
            work: list[tuple[Any, CommentedMap | CommentedSeq | None, Any | None]] = [(obj, None, None)]
        else:
            root_out = CommentedSeq()
            work = [(obj, None, None)]

       

        # Map from input container id to output container to avoid recreating
        created: dict[int, CommentedMap | CommentedSeq] = {id(obj): root_out}


        while work:
            in_node, out_parent, out_key = work.pop()

            if isinstance(in_node, dict):
                out_container = created.get(id(in_node))
                if out_container is None:
                    out_container = CommentedMap()
                    created[id(in_node)] = out_container
                    assign(out_parent, out_key, out_container)
                else:
                    # Root case: already created and assigned
                    assign(out_parent, out_key, out_container)

                # Push children in reverse to process first child next (DFS)
                items = list(in_node.items())
                for k, v in reversed(items):
                    if isinstance(v, (dict, list)):
                        # Create child container placeholder now for correct parent linkage
                        child_container = CommentedMap() if isinstance(v, dict) else CommentedSeq()
                        created[id(v)] = child_container
                        work.append((v, out_container, k))
                    else:
                        # Scalar, assign directly
                        out_container[k] = v

            elif isinstance(in_node, list):
                out_container = created.get(id(in_node))
                if out_container is None:
                    out_container = CommentedSeq()
                    created[id(in_node)] = out_container
                    assign(out_parent, out_key, out_container)
                else:
                    assign(out_parent, out_key, out_container)

                # Push children in reverse order
                for idx in reversed(range(len(in_node))):
                    v = in_node[idx]
                    if isinstance(v, (dict, list)):
                        child_container = CommentedMap() if isinstance(v, dict) else CommentedSeq()
                        created[id(v)] = child_container
                        work.append((v, out_container, idx))
                    else:
                        out_container.append(v)

            else:
                # Scalar node
                assign(out_parent, out_key, in_node)

        return root_out

    def _resolve_template_string(self, template: str):
        variables: list[tuple[str, str]] = []
        for match in self._sub_pattern.finditer(template):
            variables.append((
                match.group(0),
                self._sub_inner_text_pattern.sub('', match.group(0)),
            ))

        return variables

    def _resolve_sub_ref_queries(
        self,
        variables: list[tuple[str, str]],
        source_string: str,
    ):
        for variable, accessor in variables:
            if val := self._references.get(accessor):
                source_string = source_string.replace(variable, val)

        return source_string


from cfn_check import Collection, Rule
from example.test_models.models import (
    Resource,
    Lambda,
    LoggingGroup,
    EC2Instance
)

class ValidateResourceType(Collection):

    @Rule("Resources.*", "It checks Resource::Type is correctly definined")
    def validate_test(self, value: Resource):
        assert isinstance(value, Resource), "Not a valid CloudFormation Resource"

    @Rule("Resources.*.(Type in AWS::Lambda::Function,AWS::Serverless::Function)", "It validates a lambda is configured correctly")
    def validate_lambda(self, lambda_resource: Lambda):
        assert isinstance(lambda_resource, Lambda), "Not a valid Lambda"
        
        log_groups = self.query(
            f"Resources.{lambda_resource.Properties.LoggingConfig.LogGroup}",
            transforms=[
                lambda data: LoggingGroup(**data)
            ]
        )

        assert log_groups is not None, f"No resources found for LoggingGroup {lambda_resource.Properties.LoggingConfig.LogGroup}"
        assert len(log_groups) > 0, "No matching logging group found in Resources for Lambda"
        
    @Rule("Resources.*.(Type == AWS::Logs::LogGroup)", "It validates a logging group is configured correctly")
    def validate_logging_group(self, logging_group: LoggingGroup):
        assert isinstance(logging_group, LoggingGroup), "Not a valid logging group"

    @Rule("Resources.*.(Type == AWS::EC2::Instance)", "It validates an EC2 instance is configured correctly")
    def validate_ec2_instances(self, ec2_instance: EC2Instance):
        assert isinstance(ec2_instance, EC2Instance)
